/*
 * Foreach.cs --
 *
 * Copyright (c) 2007-2012 by Joe Mistachkin.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution of
 * this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 * RCS: @(#) $Id: $
 */

using System;
using System.Collections.Generic;
using Eagle._Attributes;
using Eagle._Components.Private;
using Eagle._Components.Public;
using Eagle._Constants;
using Eagle._Containers.Public;
using Eagle._Interfaces.Public;

namespace Eagle._Commands
{
    [ObjectId("7aa801c2-9179-4726-a536-704063349abd")]
    [CommandFlags(CommandFlags.Safe | CommandFlags.Standard)]
    [ObjectGroup("loop")]
    internal sealed class Foreach : Core
    {
        public Foreach(
            ICommandData commandData
            )
            : base(commandData)
        {
            // do nothing.
        }

        #region IExecute Members
        public override ReturnCode Execute(
            Interpreter interpreter,
            IClientData clientData,
            ArgumentList arguments,
            ref Result result
            )
        {
            ReturnCode code = ReturnCode.Ok;

            if (interpreter != null)
            {
                if (arguments != null)
                {
                    if ((arguments.Count >= 4) && ((arguments.Count % 2) == 0))
                    {
                        int numLists = ((arguments.Count - 2) / 2);
                        
                        IntList valueIndexes = new IntList();
                        List<StringList> valueLists = new List<StringList>();
                        List<StringList> variableLists = new List<StringList>();

                        int maximumIterations = 0;
                        for (int listIndex = 0; listIndex < numLists; listIndex++)
                        {
                            int argumentIndex = 1 + (listIndex * 2);
                            StringList variableList = null;

                            code = Parser.SplitList(
                                interpreter, arguments[argumentIndex], 0, Length.Invalid,
                                false, true, ref variableList, ref result);

                            if (code != ReturnCode.Ok)
                                goto done;

                            if (variableList.Count < 1)
                            {
                                result = "foreach varlist is empty";
                                code = ReturnCode.Error;
                                goto done;
                            }

                            variableLists.Add(variableList);

                            argumentIndex = 2 + (listIndex * 2);
                            StringList valueList = null;

                            code = Parser.SplitList(
                                interpreter, arguments[argumentIndex], 0, Length.Invalid,
                                false, true, ref valueList, ref result);

                            if (code != ReturnCode.Ok)
                                goto done;

                            valueLists.Add(valueList);
                            valueIndexes.Add(0);
                            
                            int iterations = valueList.Count / variableList.Count;

                            if ((valueList.Count % variableList.Count) != 0)
                                iterations++;

                            if (iterations > maximumIterations)
                                maximumIterations = iterations;
                        }

                        string body = arguments[arguments.Count - 1];
                        IScriptLocation location = arguments[arguments.Count - 1];

                        for (int iteration = 0; iteration < maximumIterations; iteration++)
                        {
                            for (int listIndex = 0; listIndex < numLists; listIndex++)
                            {
                                for (int variableIndex = 0; variableIndex < variableLists[listIndex].Count; variableIndex++)
                                {
                                    int valueIndex = valueIndexes[listIndex]++;
                                    string value = String.Empty;

                                    if (valueIndex < valueLists[listIndex].Count)
                                        value = valueLists[listIndex][valueIndex];

                                    code = interpreter.SetVariableValue(
                                        VariableFlags.None, variableLists[listIndex][variableIndex], value, null, ref result);

                                    if (code != ReturnCode.Ok)
                                    {
                                        Engine.AddErrorInformation(interpreter, result,
                                            String.Format("{0}    (setting foreach loop variable \"{1}\"",
                                                Environment.NewLine, FormatOps.Ellipsis(variableLists[listIndex][variableIndex])));

                                        goto done;
                                    }
                                }
                            }

                            code = interpreter.EvaluateScript(body, location, ref result);

                            if (code != ReturnCode.Ok)
                            {
                                if (code == ReturnCode.Continue)
                                {
                                    code = ReturnCode.Ok;
                                }
                                else if (code == ReturnCode.Break)
                                {
                                    code = ReturnCode.Ok;
                                    break;
                                }
                                else if (code == ReturnCode.Error)
                                {
                                    Engine.AddErrorInformation(interpreter, result,
                                        String.Format("{0}    (\"foreach\" body line {1})",
                                            Environment.NewLine, interpreter.ErrorLine));
                                    break;
                                }
                                else
                                {
                                    break;
                                }
                            }
                        }

                        //
                        // NOTE: Upon success, clear result.
                        //
                        if (code == ReturnCode.Ok)
                            Engine.ResetResult(interpreter, ref result);
                    }
                    else
                    {
                        result = "wrong # args: should be \"foreach varList list ?varList list ...? script\"";
                        code = ReturnCode.Error;
                    }
                }
                else
                {
                    result = "invalid argument list";
                    code = ReturnCode.Error;
                }
            }
            else
            {
                result = "invalid interpreter";
                code = ReturnCode.Error;
            }

        done:
            return code;            
        }
        #endregion
    }
}
