/*
 * Namespace.cs --
 *
 * Copyright (c) 2007-2012 by Joe Mistachkin.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution of
 * this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 * RCS: @(#) $Id: $
 */

using System;
using Eagle._Attributes;
using Eagle._Components.Private;
using Eagle._Components.Public;
using Eagle._Constants;
using Eagle._Containers.Public;
using Eagle._Interfaces.Public;

namespace Eagle._Commands
{
    [ObjectId("dbf1b8e2-0eb9-4246-ba2c-cfef01861d1d")]
    [CommandFlags(CommandFlags.Safe | CommandFlags.Standard)]
    [ObjectGroup("scriptEnvironment")]
    internal sealed class Namespace : Core
    {
        public Namespace(
            ICommandData commandData
            )
            : base(commandData)
        {
            // do nothing.
        }

        #region IEnsemble Members
        private static readonly EnsembleDictionary subCommands = new EnsembleDictionary(new string[] {
            "children", "code", "current", "delete", "eval",
            "exists", "export", "forget", "import", "inscope",
            "origin", "parent", "unknown"
        });

        public override EnsembleDictionary SubCommands
        {
            get { return subCommands; }
        }
        #endregion

        #region IExecute Members
        public override ReturnCode Execute(
            Interpreter interpreter,
            IClientData clientData,
            ArgumentList arguments,
            ref Result result
            )
        {
            ReturnCode code;

            //
            // NOTE: *WARNING* We do NOT actually support namespaces.  This command exists for the
            //       sole purpose of improving source code compatibility with simple stand alone
            //       scripts that may simply wrap themselves in a "namespace eval" block, etc.  Any
            //       other (more involved) use may not work at all or may cause undesired and/or
            //       unpredictable results.
            //
            if (interpreter != null)
            {
                if (arguments != null)
                {
                    if (arguments.Count >= 2)
                    {
                        string subCommand = arguments[1];
                        bool tried = false;

                        code = RuntimeOps.TryExecuteSubCommandFromEnsemble(
                            this, interpreter, clientData, arguments, true,
                            false, ref subCommand, ref tried, ref result);

                        if ((code == ReturnCode.Ok) && !tried)
                        {
                            switch (subCommand)
                            {
                                case "children":
                                    {
                                        if ((arguments.Count >= 2) && (arguments.Count <= 4))
                                        {
                                            if ((arguments.Count < 3) ||
                                                RuntimeOps.IsGlobalNamespace(arguments[2]))
                                            {
                                                result = String.Empty;
                                                code = ReturnCode.Ok;
                                            }
                                            else
                                            {
                                                if (RuntimeOps.HasAbsoluteNamespace(arguments[2]))
                                                    result = String.Format("namespace \"{0}\" not found", arguments[2]);
                                                else
                                                    result = String.Format("namespace \"{0}\" not found in \"{1}\"",
                                                        arguments[2], TclVars.GlobalNamespace);

                                                code = ReturnCode.Error;
                                            }
                                        }
                                        else
                                        {
                                            result = "wrong # args: should be \"namespace children ?name? ?pattern?\"";
                                            code = ReturnCode.Error;
                                        }
                                        break;
                                    }
                                case "code":
                                    {
                                        if (arguments.Count == 3)
                                        {
                                            //
                                            // NOTE: We are always in the global namespace, fake it.
                                            //
                                            result = new StringList(
                                                TclVars.GlobalNamespace + this.Name, "inscope",
                                                TclVars.GlobalNamespace, arguments[2]);

                                            code = ReturnCode.Ok;
                                        }
                                        else
                                        {
                                            result = "wrong # args: should be \"namespace code script\"";
                                            code = ReturnCode.Error;
                                        }
                                        break;
                                    }
                                case "current":
                                    {
                                        if (arguments.Count == 2)
                                        {
                                            result = TclVars.GlobalNamespace;
                                            code = ReturnCode.Ok;
                                        }
                                        else
                                        {
                                            result = "wrong # args: should be \"namespace current\"";
                                            code = ReturnCode.Error;
                                        }
                                        break;
                                    }
                                case "delete":
                                    {
                                        if (arguments.Count >= 2)
                                        {
                                            code = ReturnCode.Ok;

                                            for (int index = 2; index < arguments.Count; index++)
                                            {
                                                if (!RuntimeOps.IsGlobalNamespace(arguments[index]))
                                                {
                                                    //
                                                    // NOTE: We only know about the global namespace; an attempt
                                                    //       to delete any other namespace is an error.
                                                    //
                                                    result = String.Format(
                                                        "unknown namespace \"{0}\" in namespace delete command",
                                                        arguments[index]);

                                                    code = ReturnCode.Error;
                                                    break;
                                                }
                                            }

                                            if (code == ReturnCode.Ok)
                                            {
                                                for (int index = 2; index < arguments.Count; /* index++ */)
                                                {
                                                    //
                                                    // NOTE: Delete the one-and-only namespace (global) now (actually,
                                                    //       as soon as the evaluation stack unwinds).
                                                    //
                                                    code = interpreter.DeleteNamespace(arguments[index], false, ref result);

                                                    //
                                                    // NOTE: Since we only know about the global namespace there
                                                    //       is no point in attempting to delete it multiple times.
                                                    //
                                                    break;
                                                }
                                            }

                                            if (code == ReturnCode.Ok)
                                                result = String.Empty;
                                        }
                                        else
                                        {
                                            result = "wrong # args: should be \"namespace delete ?name name ...?\"";
                                            code = ReturnCode.Error;
                                        }
                                        break;
                                    }
                                case "eval":
                                    {
                                        if (arguments.Count >= 4)
                                        {
                                            string name = StringList.MakeList("namespace eval", arguments[2]);

                                            ICallFrame frame = interpreter.NewLinkedCallFrame(
                                                name, null, CallFrameFlags.Namespace |
                                                CallFrameFlags.Evaluate | CallFrameFlags.Tracking);

                                            interpreter.PushCallFrame(frame);

                                            if (arguments.Count == 4)
                                                code = interpreter.EvaluateScript(arguments[3], ref result);
                                            else
                                                code = interpreter.EvaluateScript(arguments, 3, ref result);

                                            if (code == ReturnCode.Error)
                                                Engine.AddErrorInformation(interpreter, result,
                                                    String.Format("{0}    (in namespace eval \"{1}\" script line {2})",
                                                        Environment.NewLine, arguments[2], interpreter.ErrorLine));

                                            //
                                            // NOTE: Pop the original call frame that we pushed above and
                                            //       any intervening scope call frames that may be leftover
                                            //       (i.e. they were not explicitly closed).
                                            //
                                            interpreter.PopCallFrame(CallFrameFlags.Scope, true);
                                        }
                                        else
                                        {
                                            result = "wrong # args: should be \"namespace eval name arg ?arg ...?\"";
                                            code = ReturnCode.Error;
                                        }
                                        break;
                                    }
                                case "exists":
                                    {
                                        if (arguments.Count == 3)
                                        {
                                            if (RuntimeOps.IsGlobalNamespace(arguments[2]))
                                                result = true;
                                            else
                                                result = false;

                                            code = ReturnCode.Ok;
                                        }
                                        else
                                        {
                                            result = "wrong # args: should be \"namespace exists name\"";
                                            code = ReturnCode.Error;
                                        }
                                        break;
                                    }
                                case "export":
                                    {
                                        if (arguments.Count >= 2)
                                        {
                                            OptionDictionary options = new OptionDictionary(
                                                new Option[] {
                                                new Option(null, OptionFlags.None, Index.Invalid, Index.Invalid, "-clear", null),
                                            });

                                            int argumentIndex = Index.Invalid;

                                            if (arguments.Count > 2)
                                                code = interpreter.GetOptions(options, arguments, 0, 2, Index.Invalid, false, ref argumentIndex, ref result);
                                            else
                                                code = ReturnCode.Ok;

                                            //
                                            // NOTE: We do not support importing or exporting of namespace commands
                                            //       because we do not really support namespaces; therefore, we do
                                            //       nothing.
                                            //
                                            if (code == ReturnCode.Ok)
                                                result = String.Empty;
                                        }
                                        else
                                        {
                                            result = "wrong # args: should be \"namespace export ?-clear? ?pattern pattern ...?\"";
                                            code = ReturnCode.Error;
                                        }
                                        break;
                                    }
                                case "forget":
                                    {
                                        if (arguments.Count >= 2)
                                        {
                                            //
                                            // NOTE: We do not support importing or exporting of namespace commands
                                            //       because we do not really support namespaces; therefore, we do
                                            //       nothing.
                                            //
                                            result = String.Empty;
                                            code = ReturnCode.Ok;
                                        }
                                        else
                                        {
                                            result = "wrong # args: should be \"namespace forget ?pattern pattern ...?\"";
                                            code = ReturnCode.Error;
                                        }
                                        break;
                                    }
                                case "import":
                                    {
                                        if (arguments.Count >= 2)
                                        {
                                            OptionDictionary options = new OptionDictionary(
                                                new Option[] {
                                                new Option(null, OptionFlags.None, Index.Invalid, Index.Invalid, "-force", null),
                                            });

                                            int argumentIndex = Index.Invalid;

                                            if (arguments.Count > 2)
                                                code = interpreter.GetOptions(options, arguments, 0, 2, Index.Invalid, false, ref argumentIndex, ref result);
                                            else
                                                code = ReturnCode.Ok;

                                            //
                                            // NOTE: We do not support importing or exporting of namespace commands
                                            //       because we do not really support namespaces; therefore, we do
                                            //       nothing.
                                            //
                                            if (code == ReturnCode.Ok)
                                                result = String.Empty;
                                        }
                                        else
                                        {
                                            result = "wrong # args: should be \"namespace import ?-force? ?pattern pattern ...?\"";
                                            code = ReturnCode.Error;
                                        }
                                        break;
                                    }
                                case "inscope":
                                    {
                                        if (arguments.Count >= 4)
                                        {
                                            if (RuntimeOps.IsGlobalNamespace(arguments[2]))
                                            {
                                                if (arguments.Count > 4)
                                                {
                                                    IScriptLocation location = null;

#if DEBUGGER && BREAKPOINTS
                                                    code = RuntimeOps.GetScriptLocation(
                                                        arguments, 3, ref location, ref result);

                                                    if (code == ReturnCode.Ok)
#endif
                                                    {
                                                        string name = StringList.MakeList("namespace inscope", arguments[2]);

                                                        ICallFrame frame = interpreter.NewLinkedCallFrame(
                                                            name, null, CallFrameFlags.Namespace |
                                                            CallFrameFlags.InScope | CallFrameFlags.Tracking);

                                                        interpreter.PushCallFrame(frame);

                                                        StringList list = new StringList(arguments, 4);

                                                        code = interpreter.EvaluateScript(
                                                            ListOps.Concat(arguments[3], list.ToString()),
                                                            location, ref result);

                                                        if (code == ReturnCode.Error)
                                                            Engine.AddErrorInformation(interpreter, result,
                                                                String.Format("{0}    (in namespace inscope \"{1}\" script line {2})",
                                                                    Environment.NewLine, arguments[2], interpreter.ErrorLine));

                                                        //
                                                        // NOTE: Pop the original call frame that we pushed above and
                                                        //       any intervening scope call frames that may be leftover
                                                        //       (i.e. they were not explicitly closed).
                                                        //
                                                        interpreter.PopCallFrame(CallFrameFlags.Scope, true);
                                                    }
                                                }
                                                else
                                                {
                                                    string name = StringList.MakeList("namespace inscope", arguments[2]);

                                                    ICallFrame frame = interpreter.NewLinkedCallFrame(
                                                        name, null, CallFrameFlags.Namespace |
                                                        CallFrameFlags.InScope | CallFrameFlags.Tracking);

                                                    interpreter.PushCallFrame(frame);

                                                    code = interpreter.EvaluateScript(arguments[3], ref result);

                                                    if (code == ReturnCode.Error)
                                                        Engine.AddErrorInformation(interpreter, result,
                                                            String.Format("{0}    (in namespace inscope \"{1}\" script line {2})",
                                                                Environment.NewLine, arguments[2], interpreter.ErrorLine));

                                                    //
                                                    // NOTE: Pop the original call frame that we pushed above and
                                                    //       any intervening scope call frames that may be leftover
                                                    //       (i.e. they were not explicitly closed).
                                                    //
                                                    interpreter.PopCallFrame(CallFrameFlags.Scope, true);
                                                }
                                            }
                                            else
                                            {
                                                result = String.Format(
                                                    "unknown namespace \"{0}\" in inscope namespace command",
                                                    arguments[2]);

                                                code = ReturnCode.Error;
                                            }
                                        }
                                        else
                                        {
                                            result = "wrong # args: should be \"namespace inscope name arg ?arg...?\"";
                                            code = ReturnCode.Error;
                                        }
                                        break;
                                    }
                                case "origin":
                                    {
                                        if (arguments.Count == 3)
                                        {
                                            string executeName = arguments[2];
                                            IExecute execute = null;

                                            code = interpreter.GetIExecuteViaResolvers(
                                                interpreter.GetResolveEngineFlags(true),
                                                executeName, null, true, true, ref execute,
                                                ref result);

                                            if (code == ReturnCode.Ok)
                                            {
                                                result = TclVars.GlobalNamespace + executeName;
                                                code = ReturnCode.Ok;
                                            }
                                        }
                                        else
                                        {
                                            result = "wrong # args: should be \"namespace origin name\"";
                                            code = ReturnCode.Error;
                                        }
                                        break;
                                    }
                                case "parent":
                                    {
                                        if ((arguments.Count == 2) || (arguments.Count == 3))
                                        {
                                            //
                                            // NOTE: Either they did not specify a namespace argument (use current
                                            //       namespace, which is always global and always exists) or they
                                            //       specified a namespace which should be the global one; otherwise,
                                            //       an error is reported because we do not really support namespaces.
                                            //
                                            if ((arguments.Count == 2) ||
                                                RuntimeOps.IsGlobalNamespace(arguments[2]))
                                            {
                                                result = String.Empty;
                                                code = ReturnCode.Ok;
                                            }
                                            else
                                            {
                                                //
                                                // NOTE: See if they prefixed the argument with "::" to figure out
                                                //       the appropriate error message (Tcl emulation).
                                                //
                                                if (RuntimeOps.HasAbsoluteNamespace(arguments[2]))
                                                    result = String.Format("namespace \"{0}\" not found", arguments[2]);
                                                else
                                                    result = String.Format("namespace \"{0}\" not found in \"{1}\"",
                                                        arguments[2], TclVars.GlobalNamespace);

                                                code = ReturnCode.Error;
                                            }
                                        }
                                        else
                                        {
                                            result = "wrong # args: should be \"namespace parent ?name?\"";
                                            code = ReturnCode.Error;
                                        }
                                        break;
                                    }
                                case "unknown":
                                    {
                                        //
                                        // NOTE: The string is currently used as the name of the command
                                        //       or procedure to execute when an unknown command is
                                        //       encountered by the engine.
                                        //
                                        if ((arguments.Count == 2) || (arguments.Count == 3))
                                        {
                                            if (arguments.Count == 3)
                                                interpreter.Unknown = arguments[2];

                                            result = interpreter.Unknown;
                                            code = ReturnCode.Ok;
                                        }
                                        else
                                        {
                                            result = "wrong # args: should be \"namespace unknown ?script?\"";
                                            code = ReturnCode.Error;
                                        }
                                        break;
                                    }
                                default:
                                    {
                                        result = RuntimeOps.BadSubCommand(null, null, subCommand, this, null, null);
                                        code = ReturnCode.Error;
                                        break;
                                    }
                            }
                        }
                    }
                    else
                    {
                        result = "wrong # args: should be \"namespace subcommand ?arg ...?\"";
                        code = ReturnCode.Error;
                    }
                }
                else
                {
                    result = "invalid argument list";
                    code = ReturnCode.Error;
                }
            }
            else
            {
                result = "invalid interpreter";
                code = ReturnCode.Error;
            }

            return code;
        }
        #endregion
    }
}
