/*
 * Sql.cs --
 *
 * Copyright (c) 2007-2012 by Joe Mistachkin.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution of
 * this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 * RCS: @(#) $Id: $
 */

using System;
using System.Collections.Generic;
using System.Data;
using System.Data.Odbc;
using System.Data.OleDb;
using System.Data.SqlClient;
using Eagle._Attributes;
using Eagle._Components.Private;
using Eagle._Components.Public;
using Eagle._Constants;
using Eagle._Containers.Public;
using Eagle._Interfaces.Public;
using IsolationLevel = System.Data.IsolationLevel;

namespace Eagle._Commands
{
    [ObjectId("dbc78d04-325d-4805-a118-3cfeeddfb8fc")]
    [CommandFlags(CommandFlags.Unsafe
#if NATIVE && WINDOWS
        //
        // NOTE: Uses native code indirectly for profiling [sql execute] with
        //       the "-time" option (on Windows only).
        //
        | CommandFlags.NativeCode
#endif
        )]
    [ObjectGroup("managedEnvironment")]
    internal sealed class Sql : Core
    {
        private static Dictionary<DbConnectionType, string> DbConnectionTypeFullNames = null;
        private static Dictionary<DbConnectionType, string> DbConnectionTypeNames = null;

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public Sql(
            ICommandData commandData
            )
            : base(commandData)
        {
            //
            // NOTE: One-time initialization, these are not per-interpreter datums and never
            //       change.
            //
            if (DbConnectionTypeFullNames == null)
            {
                DbConnectionTypeFullNames = new Dictionary<DbConnectionType,string>();

                //
                // NOTE: This is optional because it requires the System.Data.OracleClient
                //       assembly to be loaded.
                //
                DbConnectionTypeFullNames.Add(DbConnectionType.Oracle,
                    "System.Data.OracleClient.OracleConnection, System.Data.OracleClient, " +
                    "Version=2.0.0.0, Culture=neutral, PublicKeyToken=b77a5c561934e089");

                //
                // NOTE: This is optional because it requires the .NET Framework v3.5
                //       (SP1 or higher?) to be installed.
                //
                DbConnectionTypeFullNames.Add(DbConnectionType.SqlCe,
                    "System.Data.SqlServerCe.SqlCeConnection, System.Data.SqlServerCe, " +
                    "Version=3.5.1.0, Culture=neutral, PublicKeyToken=89845dcd8080cc91");

                //
                // NOTE: This is optional because it requires the System.Data.SQLite
                //       assembly to be loaded (http://system.data.sqlite.org/ OR
                //       http://sf.net/projects/sqlite-dotnet2/).
                //
                DbConnectionTypeFullNames.Add(DbConnectionType.SQLite,
                    "System.Data.SQLite.SQLiteConnection, System.Data.SQLite, " +
                    "Version=1.0, Culture=neutral, PublicKeyToken=db937bc2d44ff139");
            }

            if (DbConnectionTypeNames == null)
            {
                DbConnectionTypeNames = new Dictionary<DbConnectionType, string>();

                //
                // NOTE: This is optional because it requires the System.Data.SQLite
                //       assembly to be loaded (http://system.data.sqlite.org/ OR
                //       http://sf.net/projects/sqlite-dotnet2/).
                //
                DbConnectionTypeNames.Add(DbConnectionType.SQLite,
                    "System.Data.SQLite.SQLiteConnection");
            }
        }

        #region IEnsemble Members
        private static readonly EnsembleDictionary subCommands = new EnsembleDictionary(new string[] {
            "close", "connection", "execute", "open", "transaction"
        });

        public override EnsembleDictionary SubCommands
        {
            get { return subCommands; }
        }
        #endregion

        #region IExecute Members
        public override ReturnCode Execute(
            Interpreter interpreter,
            IClientData clientData,
            ArgumentList arguments,
            ref Result result
            )
        {
            ReturnCode code = ReturnCode.Ok;

            if (interpreter != null)
            {
                if (arguments != null)
                {
                    if (arguments.Count >= 2)
                    {
                        string subCommand = arguments[1];
                        bool tried = false;

                        code = RuntimeOps.TryExecuteSubCommandFromEnsemble(
                            this, interpreter, clientData, arguments, true,
                            false, ref subCommand, ref tried, ref result);

                        if ((code == ReturnCode.Ok) && !tried)
                        {
                            switch (subCommand)
                            {
                                case "close":
                                    {
                                        if (arguments.Count == 3)
                                        {
                                            IDbConnection connection = null;

                                            code = interpreter.GetDbConnection(arguments[2], ref connection, ref result);

                                            if (code == ReturnCode.Ok)
                                            {
                                                //
                                                // NOTE: We are going to modify the interpreter state, make
                                                //       sure it is not set to read-only.
                                                //
                                                if (!interpreter.ReadOnly)
                                                {
                                                    if (interpreter.HasDbConnections(ref result))
                                                    {
                                                        try
                                                        {
                                                            connection.Close();

                                                            lock (interpreter.SyncRoot)
                                                            {
                                                                interpreter.connections.Remove(arguments[2]);
                                                            }

#if NOTIFY
                                                            /* IGNORED */
                                                            interpreter.CheckNotifications(
                                                                NotifyType.Connection, NotifyFlags.Removed,
                                                                connection, interpreter, null, null, null,
                                                                ref result);
#endif

                                                            connection = null;
                                                        }
                                                        catch (Exception e)
                                                        {
                                                            Engine.SetExceptionErrorCode(interpreter, e);

                                                            result = e;
                                                            code = ReturnCode.Error;
                                                        }
                                                    }
                                                    else
                                                    {
                                                        code = ReturnCode.Error;
                                                    }
                                                }
                                                else
                                                {
                                                    result = "interpreter is read-only";
                                                    code = ReturnCode.Error;
                                                }
                                            }
                                        }
                                        else
                                        {
                                            result = "wrong # args: should be \"sql close connection\"";
                                            code = ReturnCode.Error;
                                        }
                                        break;
                                    }
                                case "connection":
                                    {
                                        if (arguments.Count == 3)
                                        {
                                            if (interpreter.HasDbConnections(ref result))
                                            {
                                                IDbConnection connection = null;

                                                code = interpreter.GetDbConnection(arguments[2], ref connection, ref result);

                                                if (code == ReturnCode.Ok)
                                                {
                                                    try
                                                    {
                                                        result = StringList.MakeList(
                                                            "type", connection.GetType().Name,
                                                            "state", connection.State,
                                                            "database", connection.Database,
                                                            "timeout", connection.ConnectionTimeout,
                                                            "string", connection.ConnectionString);
                                                    }
                                                    catch (Exception e)
                                                    {
                                                        Engine.SetExceptionErrorCode(interpreter, e);

                                                        result = e;
                                                        code = ReturnCode.Error;
                                                    }
                                                }
                                            }
                                            else
                                            {
                                                code = ReturnCode.Error;
                                            }
                                        }
                                        else
                                        {
                                            result = "sql connection connection";
                                            code = ReturnCode.Error;
                                        }
                                        break;
                                    }
                                case "execute":
                                    {
                                        if (arguments.Count >= 4)
                                        {
                                            OptionDictionary options = new OptionDictionary(
                                                new Option[] {
                                                new Option(null, OptionFlags.None, Index.Invalid, Index.Invalid, "-verbatim", null),
                                                new Option(typeof(DateTimeBehavior), OptionFlags.MustHaveEnumValue, Index.Invalid, Index.Invalid,
                                                    "-datetimebehavior", new Variant(DateTimeBehavior.Default)),
                                                new Option(typeof(ValueFlags), OptionFlags.MustHaveEnumValue, Index.Invalid, Index.Invalid,
                                                    "-valueflags", new Variant(ValueFlags.AnyNonCharacter)),
                                                new Option(null, OptionFlags.MustHaveValue, Index.Invalid, Index.Invalid, "-datetimeformat", null),
                                                new Option(null, OptionFlags.MustHaveValue, Index.Invalid, Index.Invalid, "-transaction", null),
                                                new Option(null, OptionFlags.MustHaveValue, Index.Invalid, Index.Invalid, "-rowsvar", null),
                                                new Option(null, OptionFlags.None, Index.Invalid, Index.Invalid, "-time", null),
                                                new Option(null, OptionFlags.MustHaveBooleanValue, Index.Invalid, Index.Invalid, "-allownull", null),
                                                new Option(null, OptionFlags.MustHaveBooleanValue, Index.Invalid, Index.Invalid, "-pairs", null),
                                                new Option(null, OptionFlags.MustHaveBooleanValue, Index.Invalid, Index.Invalid, "-names", null),
                                                new Option(null, OptionFlags.MustHaveValue, Index.Invalid, Index.Invalid, "-timevar", null),
                                                new Option(null, OptionFlags.MustHaveIntegerValue, Index.Invalid, Index.Invalid, "-timeout", null),
                                                new Option(null, OptionFlags.MustHaveIntegerValue, Index.Invalid, Index.Invalid, "-limit", null),
                                                new Option(typeof(CommandType), OptionFlags.MustHaveEnumValue, Index.Invalid, Index.Invalid, "-commandtype", null),
                                                new Option(typeof(DbResultFormat), OptionFlags.MustHaveEnumValue, Index.Invalid, Index.Invalid, "-format", null),
                                                new Option(typeof(DbExecuteType), OptionFlags.MustHaveEnumValue, Index.Invalid, Index.Invalid, "-execute", null),
                                                new Option(typeof(CommandBehavior), OptionFlags.MustHaveEnumValue, Index.Invalid, Index.Invalid, "-behavior",
                                                    new Variant(CommandBehavior.Default)),
                                                new Option(null, OptionFlags.MustHaveValue, Index.Invalid, Index.Invalid, "-objectname", null),
                                                new Option(null, OptionFlags.MustHaveTypeValue, Index.Invalid, Index.Invalid, "-objecttype", null),
                                                new Option(typeof(ObjectFlags), OptionFlags.MustHaveEnumValue, Index.Invalid, Index.Invalid, "-objectflags",
                                                    new Variant(ObjectFlags.Default)),
                                                new Option(null, OptionFlags.None, Index.Invalid, Index.Invalid, "-nocreate", null),
                                                new Option(null, OptionFlags.None, Index.Invalid, Index.Invalid, "-nodispose", null),
                                                new Option(null, OptionFlags.None, Index.Invalid, Index.Invalid, "-alias", null),
                                                new Option(null, OptionFlags.None, Index.Invalid, Index.Invalid, "-aliasall", null),
#if NATIVE && TCL
                                                new Option(null, OptionFlags.MustHaveTclInterpreterValue, Index.Invalid, Index.Invalid, "-tcl", null),
#endif
                                                new Option(null, OptionFlags.None, Index.Invalid, Index.Invalid, "-noforcedelete", null),
                                                new Option(null, OptionFlags.None, Index.Invalid, Index.Invalid, "-tostring", null),
                                                new Option(null, OptionFlags.None, Index.Invalid, Index.Invalid, Option.EndOfOptions, null)
                                            });

                                            int argumentIndex = Index.Invalid;

                                            code = interpreter.GetOptions(options, arguments, 0, 2, Index.Invalid, false, ref argumentIndex, ref result);

                                            if (code == ReturnCode.Ok)
                                            {
                                                if ((argumentIndex != Index.Invalid) && ((argumentIndex + 1) < arguments.Count))
                                                {
                                                    Variant value = null;
                                                    IDbTransaction transaction = null;

                                                    if (options.IsPresent("-transaction", ref value))
                                                    {
                                                        string transactionName = value.ToString();

                                                        if (!String.IsNullOrEmpty(transactionName))
                                                        {
                                                            if (interpreter.HasDbTransactions(ref result))
                                                                code = interpreter.GetDbTransaction(
                                                                    transactionName, ref transaction, ref result);
                                                            else
                                                                code = ReturnCode.Error;
                                                        }
                                                    }

                                                    if (code == ReturnCode.Ok)
                                                    {
                                                        CommandBehavior commandBehavior = CommandBehavior.Default;

                                                        if (options.IsPresent("-behavior", ref value))
                                                            commandBehavior = (CommandBehavior)value.Value;

                                                        CommandType commandType = CommandType.Text; /* TODO: Good default? */

                                                        if (options.IsPresent("-commandtype", ref value))
                                                            commandType = (CommandType)value.Value;

                                                        DbExecuteType executeType = DbExecuteType.Default;

                                                        if (options.IsPresent("-execute", ref value))
                                                            executeType = (DbExecuteType)value.Value;

                                                        DbResultFormat resultFormat = DbResultFormat.Default;

                                                        if (options.IsPresent("-format", ref value))
                                                            resultFormat = (DbResultFormat)value.Value;

                                                        int commandTimeout = _Timeout.Infinite; /* TODO: Good default? */

                                                        if (options.IsPresent("-timeout", ref value))
                                                            commandTimeout = (int)value.Value;

                                                        int limit = 0;

                                                        if (options.IsPresent("-limit", ref value))
                                                            limit = (int)value.Value;

                                                        string rowsVarName = Vars.ResultSet.Rows;

                                                        if (options.IsPresent("-rowsvar", ref value))
                                                            rowsVarName = value.ToString();

                                                        bool allowNull = false; /* TODO: Good default? */

                                                        if (options.IsPresent("-allownull", ref value))
                                                            allowNull = (bool)value.Value;

                                                        bool pairs = true; /* TODO: Good default? */

                                                        if (options.IsPresent("-pairs", ref value))
                                                            pairs = (bool)value.Value;

                                                        bool names = true; /* TODO: Good default? */

                                                        if (options.IsPresent("-names", ref value))
                                                            names = (bool)value.Value;

                                                        bool time = false;

                                                        if (options.IsPresent("-time"))
                                                            time = true;

                                                        string timeVarName = Vars.ResultSet.Time;

                                                        if (options.IsPresent("-timevar", ref value))
                                                            timeVarName = value.ToString();

                                                        ValueFlags valueFlags = ValueFlags.AnyNonCharacter;

                                                        if (options.IsPresent("-valueflags", ref value))
                                                            valueFlags = (ValueFlags)value.Value;

                                                        DateTimeBehavior dateTimeBehavior = DateTimeBehavior.Default;

                                                        if (options.IsPresent("-datetimebehavior", ref value))
                                                            dateTimeBehavior = (DateTimeBehavior)value.Value;

                                                        string dateTimeFormat = null;

                                                        if (options.IsPresent("-datetimeformat", ref value))
                                                            dateTimeFormat = value.ToString();

                                                        bool verbatim = false;

                                                        if (options.IsPresent("-verbatim"))
                                                            verbatim = true;

                                                        bool create = true;

                                                        if (options.IsPresent("-nocreate"))
                                                            create = false;

                                                        bool dispose = true;

                                                        if (options.IsPresent("-nodispose"))
                                                            dispose = false;

                                                        bool alias = false;

                                                        if (options.IsPresent("-alias"))
                                                            alias = true;

                                                        bool aliasAll = false;

                                                        if (options.IsPresent("-aliasall"))
                                                            aliasAll = true;

                                                        bool toString = false;

                                                        if (options.IsPresent("-tostring"))
                                                            toString = true;

                                                        Type type = null;

                                                        if (options.IsPresent("-objecttype", ref value))
                                                            type = (Type)value.Value;

                                                        string interpName = null;

#if NATIVE && TCL
                                                        if (options.IsPresent("-tcl", ref value))
                                                            interpName = value.ToString();
#endif

                                                        ObjectFlags objectFlags = ObjectFlags.Default;

                                                        if (options.IsPresent("-objectflags", ref value))
                                                            objectFlags = (ObjectFlags)value.Value;

                                                        if (options.IsPresent("-noforcedelete"))
                                                            objectFlags &= ~ObjectFlags.ForceDelete;

                                                        string objectName = null;

                                                        if (options.IsPresent("-objectname", ref value))
                                                            objectName = value.ToString();

                                                        if (interpreter.HasDbConnections(ref result))
                                                        {
                                                            IDbConnection connection = null;

                                                            code = interpreter.GetDbConnection(
                                                                arguments[argumentIndex], ref connection, ref result);

                                                            if (code == ReturnCode.Ok)
                                                            {
                                                                IDbCommand command = null;

                                                                try
                                                                {
                                                                    command = connection.CreateCommand();

                                                                    //
                                                                    // NOTE: Set the command text itself to the value of the second
                                                                    //       arguemnt after the options.
                                                                    //
                                                                    command.CommandText = arguments[argumentIndex + 1];

                                                                    //
                                                                    // NOTE: If the timeout was supplied, set the timeout value now;
                                                                    //       otherwise, leave it alone to retain the default for the
                                                                    //       underlying provider.
                                                                    //
                                                                    if (commandTimeout != _Timeout.Infinite)
                                                                        command.CommandTimeout = commandTimeout;

                                                                    //
                                                                    // NOTE: Set the command type to the value specified in the option
                                                                    //       (or the default if none was supplied).
                                                                    //
                                                                    command.CommandType = commandType;

                                                                    //
                                                                    // NOTE: Setup the transaction for this query.  If this is set
                                                                    //       to null, default transaction semantics may be used by
                                                                    //       the underlying data provider.
                                                                    //
                                                                    command.Transaction = transaction; /* throw */

                                                                    //
                                                                    // NOTE: Add any supplied parameters to this command.
                                                                    //
                                                                    for (argumentIndex += 2; argumentIndex < arguments.Count; argumentIndex++)
                                                                    {
                                                                        StringList parameterList = null;

                                                                        code = Parser.SplitList(
                                                                            interpreter, arguments[argumentIndex], 0, Length.Invalid,
                                                                            false, true, ref parameterList, ref result);

                                                                        if (code != ReturnCode.Ok)
                                                                            break;

                                                                        if (parameterList.Count >= 1)
                                                                        {
                                                                            IDbDataParameter parameter = command.CreateParameter();

                                                                            parameter.ParameterName = parameterList[0];

                                                                            if ((parameterList.Count >= 2) &&
                                                                                !String.IsNullOrEmpty(parameterList[1]))
                                                                            {
                                                                                object enumValue = EnumOps.TryParseEnum(
                                                                                    typeof(DbType), parameterList[1], true);

                                                                                if (enumValue is DbType)
                                                                                {
                                                                                    parameter.DbType = (DbType)enumValue;
                                                                                }
                                                                                else
                                                                                {
                                                                                    result = RuntimeOps.BadValue(
                                                                                        null, "database type", parameterList[1],
                                                                                        Enum.GetNames(typeof(DbType)), null, null);

                                                                                    code = ReturnCode.Error;
                                                                                    break;
                                                                                }
                                                                            }

                                                                            if (parameterList.Count >= 3)
                                                                            {
                                                                                object parameterValue = parameterList[2];

                                                                                if (parameterValue is string)
                                                                                    /* IGNORED */
                                                                                    Value.GetObject(interpreter, (string)parameterValue,
                                                                                        ref parameterValue);

                                                                                if (!verbatim && (parameterValue is string))
                                                                                {
                                                                                    ValueFlags parameterValueFlags = valueFlags;

                                                                                    if (parameterList.Count >= 5)
                                                                                    {
                                                                                        object enumValue = EnumOps.TryParseFlagsEnum(
                                                                                            interpreter, typeof(ValueFlags),
                                                                                            parameterValueFlags.ToString(), parameterList[4],
                                                                                            interpreter.CultureInfo, true, ref result);

                                                                                        if (enumValue is ValueFlags)
                                                                                        {
                                                                                            parameterValueFlags = (ValueFlags)enumValue;
                                                                                        }
                                                                                        else
                                                                                        {
                                                                                            result = RuntimeOps.BadValue(
                                                                                                null, "value flags", parameterList[4],
                                                                                                Enum.GetNames(typeof(ValueFlags)), null,
                                                                                                null);

                                                                                            code = ReturnCode.Error;
                                                                                            break;
                                                                                        }
                                                                                    }

                                                                                    /* IGNORED */
                                                                                    Value.GetValue((string)parameterValue,
                                                                                        parameterValueFlags, true, interpreter.CultureInfo,
                                                                                        ref parameterValue);
                                                                                }

                                                                                parameter.Value = parameterValue;
                                                                            }
                                                                            else
                                                                            {
                                                                                parameter.Value = DBNull.Value;
                                                                            }

                                                                            if ((parameterList.Count >= 4) &&
                                                                                !String.IsNullOrEmpty(parameterList[3]))
                                                                            {
                                                                                int size = 0;

                                                                                code = Value.GetInteger2(parameterList[3], ValueFlags.AnyInteger,
                                                                                    interpreter.CultureInfo, ref size, ref result);

                                                                                if (code != ReturnCode.Ok)
                                                                                    break;

                                                                                parameter.Size = size;
                                                                            }

                                                                            command.Parameters.Add(parameter);
                                                                        }
                                                                        else
                                                                        {
                                                                            result = "parameter missing required element \"name\"";
                                                                            code = ReturnCode.Error;
                                                                            break;
                                                                        }
                                                                    }

                                                                    //
                                                                    // NOTE: Make sure we succeeded parsing the optional parameters,
                                                                    //       if any were provided.
                                                                    //
                                                                    if (code == ReturnCode.Ok)
                                                                    {
                                                                        //
                                                                        // NOTE: These variables are used to measure performance if the
                                                                        //       -time option is enabled.
                                                                        //
                                                                        long startCount = 0;
                                                                        long stopCount = 0;
                                                                        double microseconds = 0;

                                                                        //
                                                                        // NOTE: Always prepare the statement, even though it may result
                                                                        //       in a no-op.
                                                                        //
                                                                        if (time)
                                                                            RuntimeOps.StartProfiling(ref startCount, ref microseconds);

                                                                        command.Prepare();

                                                                        if (time)
                                                                        {
                                                                            RuntimeOps.StopProfiling(
                                                                                startCount, ref stopCount, ref microseconds);

                                                                            ReturnCode setCode;
                                                                            Result setError = null;

                                                                            setCode = interpreter.SetVariableValue2(
                                                                                VariableFlags.None, timeVarName,
                                                                                Vars.ResultSet.Prepare, microseconds.ToString(),
                                                                                null, ref setError);

                                                                            if (setCode != ReturnCode.Ok)
                                                                                DebugOps.Complain(interpreter, setCode, setError);

                                                                            RuntimeOps.StartProfiling(ref startCount, ref microseconds);
                                                                        }

                                                                        //
                                                                        // NOTE: Execute the appropriate method for this execution type.
                                                                        //
                                                                        switch (executeType)
                                                                        {
                                                                            case DbExecuteType.None:
                                                                                {
                                                                                    result = String.Empty;
                                                                                    break;
                                                                                }
                                                                            case DbExecuteType.NonQuery:
                                                                                {
                                                                                    try
                                                                                    {
                                                                                        result = command.ExecuteNonQuery();
                                                                                    }
                                                                                    catch (Exception e)
                                                                                    {
                                                                                        Engine.SetExceptionErrorCode(interpreter, e);

                                                                                        result = e;
                                                                                        code = ReturnCode.Error;
                                                                                    }
                                                                                    break;
                                                                                }
                                                                            case DbExecuteType.Scalar:
                                                                                {
                                                                                    try
                                                                                    {
                                                                                        result = MarshalOps.FixupDataValue(
                                                                                            command.ExecuteScalar(), dateTimeBehavior,
                                                                                            dateTimeFormat, null);
                                                                                    }
                                                                                    catch (Exception e)
                                                                                    {
                                                                                        Engine.SetExceptionErrorCode(interpreter, e);

                                                                                        result = e;
                                                                                        code = ReturnCode.Error;
                                                                                    }
                                                                                    break;
                                                                                }
                                                                            case DbExecuteType.Reader:
                                                                                {
                                                                                    bool close = true;
                                                                                    IDataReader reader = null;

                                                                                    try
                                                                                    {
                                                                                        reader = command.ExecuteReader(
                                                                                            commandBehavior);

                                                                                        switch (resultFormat)
                                                                                        {
                                                                                            case DbResultFormat.None:
                                                                                                {
                                                                                                    result = String.Empty;
                                                                                                    code = ReturnCode.Ok;
                                                                                                    break;
                                                                                                }
                                                                                            case DbResultFormat.RawArray:
                                                                                                {
                                                                                                    code = RuntimeOps.DataReaderToArray(
                                                                                                        interpreter, reader, rowsVarName,
                                                                                                        dateTimeBehavior, dateTimeFormat,
                                                                                                        limit, allowNull, pairs, names,
                                                                                                        ref result);

                                                                                                    if (code == ReturnCode.Ok)
                                                                                                        result = String.Empty;

                                                                                                    break;
                                                                                                }
                                                                                            case DbResultFormat.RawList:
                                                                                                {
                                                                                                    StringList list = null;

                                                                                                    code = RuntimeOps.DataReaderToList(
                                                                                                        interpreter, reader, dateTimeBehavior,
                                                                                                        dateTimeFormat, limit, allowNull, pairs,
                                                                                                        names, ref list, ref result);

                                                                                                    if (code == ReturnCode.Ok)
                                                                                                        result = list;

                                                                                                    break;
                                                                                                }
                                                                                            case DbResultFormat.Array:
                                                                                                {
                                                                                                    pairs = true;
                                                                                                    names = true;

                                                                                                    goto case DbResultFormat.RawArray;
                                                                                                }
                                                                                            case DbResultFormat.List:
                                                                                                {
                                                                                                    pairs = false;
                                                                                                    names = false;

                                                                                                    goto case DbResultFormat.RawList;
                                                                                                }
                                                                                            case DbResultFormat.Dictionary:
                                                                                                {
                                                                                                    pairs = false;
                                                                                                    names = true;

                                                                                                    goto case DbResultFormat.RawList;
                                                                                                }
                                                                                            case DbResultFormat.DataReader:
                                                                                                {
                                                                                                    OptionDictionary readerOptions =
                                                                                                        RuntimeOps.GetInvokeOptions(aliasAll);

                                                                                                    code = MarshalOps.FixupReturnValue(
                                                                                                        interpreter, interpreter.Binder,
                                                                                                        interpreter.CultureInfo, type,
                                                                                                        objectFlags, readerOptions,
                                                                                                        objectName, interpName, reader,
                                                                                                        create, dispose, alias, aliasAll,
                                                                                                        toString, ref result);

                                                                                                    if ((code == ReturnCode.Ok) &&
                                                                                                        (interpreter.GetObject(reader,
                                                                                                            true, true) == ReturnCode.Ok))
                                                                                                    {
                                                                                                        //
                                                                                                        // NOTE: The IDataReader object was
                                                                                                        //       successfully transferred to
                                                                                                        //       the interpreter object list;
                                                                                                        //       therefore, we no longer need
                                                                                                        //       or want to close it.
                                                                                                        //
                                                                                                        close = false;
                                                                                                    }
                                                                                                    break;
                                                                                                }
                                                                                            default:
                                                                                                {
                                                                                                    result = String.Format(
                                                                                                        "unsupported result format \"{0}\"",
                                                                                                        resultFormat);

                                                                                                    code = ReturnCode.Error;
                                                                                                    break;
                                                                                                }
                                                                                        }
                                                                                    }
                                                                                    catch (Exception e)
                                                                                    {
                                                                                        Engine.SetExceptionErrorCode(interpreter, e);

                                                                                        result = e;
                                                                                        code = ReturnCode.Error;
                                                                                    }
                                                                                    finally
                                                                                    {
                                                                                        if (reader != null)
                                                                                        {
                                                                                            if (close)
                                                                                                reader.Close();

                                                                                            reader = null;
                                                                                        }
                                                                                    }
                                                                                    break;
                                                                                }
                                                                            default:
                                                                                {
                                                                                    result = String.Format(
                                                                                        "unsupported execution type \"{0}\"",
                                                                                        executeType);

                                                                                    code = ReturnCode.Error;
                                                                                    break;
                                                                                }
                                                                        }

                                                                        if (time)
                                                                        {
                                                                            RuntimeOps.StopProfiling(
                                                                                startCount, ref stopCount, ref microseconds);

                                                                            ReturnCode setCode;
                                                                            Result setError = null;

                                                                            setCode = interpreter.SetVariableValue2(
                                                                                VariableFlags.None, timeVarName,
                                                                                Vars.ResultSet.Execute, microseconds.ToString(),
                                                                                null, ref setError);

                                                                            if (setCode != ReturnCode.Ok)
                                                                                DebugOps.Complain(interpreter, setCode, setError);
                                                                        }
                                                                    }
                                                                }
                                                                catch (Exception e)
                                                                {
                                                                    Engine.SetExceptionErrorCode(interpreter, e);

                                                                    result = e;
                                                                    code = ReturnCode.Error;
                                                                }
                                                                finally
                                                                {
                                                                    if (command != null)
                                                                    {
                                                                        command.Dispose();
                                                                        command = null;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else
                                                        {
                                                            code = ReturnCode.Error;
                                                        }
                                                    }
                                                }
                                                else
                                                {
                                                    if ((argumentIndex != Index.Invalid) &&
                                                        Option.LooksLikeOption(arguments[argumentIndex]))
                                                    {
                                                        result = OptionDictionary.BadOption(options, arguments[argumentIndex]);
                                                    }
                                                    else
                                                    {
                                                        result = "wrong # args: should be \"sql execute ?options? connection string ?{paramName ?paramType? ?paramValue? ?paramSize? ?paramValueFlags?} ...?\"";
                                                    }

                                                    code = ReturnCode.Error;
                                                }
                                            }
                                        }
                                        else
                                        {
                                            result = "wrong # args: should be \"sql execute ?options? connection string ?{paramName ?paramType? ?paramValue? ?paramSize? ?paramValueFlags?} ...?\"";
                                            code = ReturnCode.Error;
                                        }
                                        break;
                                    }
                                case "open":
                                    {
                                        if (arguments.Count >= 3)
                                        {
                                            OptionDictionary options = new OptionDictionary(
                                                new Option[] {
                                                new Option(typeof(DbConnectionType), OptionFlags.MustHaveEnumValue, Index.Invalid, Index.Invalid, "-type", null),
                                                new Option(null, OptionFlags.MustHaveValue, Index.Invalid, Index.Invalid, "-typename", null),
                                                new Option(null, OptionFlags.MustHaveValue, Index.Invalid, Index.Invalid, "-typefullname", null),
                                                new Option(null, OptionFlags.None, Index.Invalid, Index.Invalid, "-stricttype", null),
                                                new Option(null, OptionFlags.None, Index.Invalid, Index.Invalid, "-verbose", null),
                                                new Option(null, OptionFlags.None, Index.Invalid, Index.Invalid, "-nocase", null),
                                                new Option(null, OptionFlags.None, Index.Invalid, Index.Invalid, Option.EndOfOptions, null)
                                            });

                                            int argumentIndex = Index.Invalid;

                                            code = interpreter.GetOptions(options, arguments, 0, 2, Index.Invalid, false, ref argumentIndex, ref result);

                                            if (code == ReturnCode.Ok)
                                            {
                                                if ((argumentIndex != Index.Invalid) && ((argumentIndex + 1) == arguments.Count))
                                                {
                                                    Variant value = null;
                                                    DbConnectionType dbConnectionType = DbConnectionType.Sql; /* TODO: Good default? */

                                                    if (options.IsPresent("-type", ref value))
                                                        dbConnectionType = (DbConnectionType)value.Value;

                                                    string typeName = null;

                                                    if (options.IsPresent("-typename", ref value))
                                                        typeName = value.ToString();

                                                    string typeFullName = null;

                                                    if (options.IsPresent("-typefullname", ref value))
                                                        typeFullName = value.ToString();

                                                    //
                                                    // NOTE: Perform a case-insensitive search for the type name?
                                                    //
                                                    bool noCase = false;

                                                    if (options.IsPresent("-nocase"))
                                                        noCase = true;

                                                    //
                                                    // NOTE: Prevent any magical type searches (i.e. use their specified
                                                    //       type string verbatim)?
                                                    //
                                                    bool strictType = false;

                                                    if (options.IsPresent("-stricttype"))
                                                        strictType = true;

                                                    //
                                                    // NOTE: Return all Type exception information (this can be very
                                                    //       costly for performance).
                                                    //
                                                    bool verbose = false;

                                                    if (options.IsPresent("-verbose"))
                                                        verbose = true;

                                                    //
                                                    // NOTE: We are going to modify the interpreter state, make
                                                    //       sure it is not set to read-only.
                                                    //
                                                    if (!interpreter.ReadOnly)
                                                    {
                                                        if (interpreter.HasDbConnections(ref result))
                                                        {
                                                            try
                                                            {
                                                                IDbConnection connection = null;

                                                                switch (dbConnectionType)
                                                                {
                                                                    case DbConnectionType.None:
                                                                        connection = null;
                                                                        break;
                                                                    case DbConnectionType.Odbc:
                                                                        connection = new OdbcConnection(arguments[argumentIndex]);
                                                                        break;
                                                                    case DbConnectionType.OleDb:
                                                                        connection = new OleDbConnection(arguments[argumentIndex]);
                                                                        break;
                                                                    case DbConnectionType.Sql:
                                                                        connection = new SqlConnection(arguments[argumentIndex]);
                                                                        break;
                                                                    case DbConnectionType.Other:
                                                                        {
                                                                            if (!String.IsNullOrEmpty(typeFullName) || !String.IsNullOrEmpty(typeName))
                                                                            {
                                                                                Type type = null;
                                                                                ResultList errors = null;

                                                                                if (!String.IsNullOrEmpty(typeFullName))
                                                                                {
                                                                                    code = Value.GetType(
                                                                                        interpreter, typeFullName, interpreter.GetAppDomain(),
                                                                                        strictType, verbose, noCase, interpreter.CultureInfo,
                                                                                        ref type, ref errors);
                                                                                }

                                                                                if (((code != ReturnCode.Ok) || String.IsNullOrEmpty(typeFullName)) &&
                                                                                    !String.IsNullOrEmpty(typeName))
                                                                                {
                                                                                    code = Value.GetType(
                                                                                        interpreter, typeName, interpreter.GetAppDomain(),
                                                                                        strictType, verbose, noCase, interpreter.CultureInfo,
                                                                                        ref type, ref errors);
                                                                                }

                                                                                if (code == ReturnCode.Ok)
                                                                                {
                                                                                    connection = (IDbConnection)Activator.CreateInstance(
                                                                                        type, new object[] { arguments[argumentIndex].ToString() });
                                                                                }
                                                                                else
                                                                                {
                                                                                    result = errors;
                                                                                }
                                                                            }
                                                                            else
                                                                            {
                                                                                result = String.Format(
                                                                                    "invalid type name for database connection type \"{0}\"",
                                                                                    dbConnectionType);

                                                                                code = ReturnCode.Error;
                                                                            }
                                                                            break;
                                                                        }
                                                                    default:
                                                                        {
                                                                            //
                                                                            // NOTE: Lookup the type name and/or full name and then go to the "other"
                                                                            //       case (for database connection types that are not "built-in").
                                                                            //
                                                                            bool found = false;

                                                                            if ((DbConnectionTypeFullNames != null) &&
                                                                                DbConnectionTypeFullNames.TryGetValue(dbConnectionType, out typeFullName))
                                                                            {
                                                                                found = true;
                                                                            }

                                                                            if ((DbConnectionTypeNames != null) &&
                                                                                DbConnectionTypeNames.TryGetValue(dbConnectionType, out typeName))
                                                                            {
                                                                                found = true;
                                                                            }

                                                                            if (found)
                                                                            {
                                                                                goto case DbConnectionType.Other;
                                                                            }
                                                                            else
                                                                            {
                                                                                result = String.Format(
                                                                                    "unsupported database connection type \"{0}\"",
                                                                                    dbConnectionType);

                                                                                code = ReturnCode.Error;
                                                                            }
                                                                            break;
                                                                        }
                                                                }

                                                                if (code == ReturnCode.Ok)
                                                                {
                                                                    if (connection != null)
                                                                        connection.Open();

                                                                    result = FormatOps.DatabaseObjectName(connection,
                                                                        dbConnectionType.ToString() + "Connection",
                                                                        interpreter.NextId());

                                                                    lock (interpreter.SyncRoot)
                                                                    {
                                                                        interpreter.connections.Add(result, connection);
                                                                    }

#if NOTIFY
                                                                    /* IGNORED */
                                                                    interpreter.CheckNotifications(
                                                                        NotifyType.Connection, NotifyFlags.Added,
                                                                        connection, interpreter, null, null, null,
                                                                        ref result);
#endif
                                                                }
                                                            }
                                                            catch (Exception e)
                                                            {
                                                                Engine.SetExceptionErrorCode(interpreter, e);

                                                                result = e;
                                                                code = ReturnCode.Error;
                                                            }
                                                        }
                                                        else
                                                        {
                                                            code = ReturnCode.Error;
                                                        }
                                                    }
                                                    else
                                                    {
                                                        result = "interpreter is read-only";
                                                        code = ReturnCode.Error;
                                                    }
                                                }
                                                else
                                                {
                                                    if ((argumentIndex != Index.Invalid) &&
                                                        Option.LooksLikeOption(arguments[argumentIndex]))
                                                    {
                                                        result = OptionDictionary.BadOption(options, arguments[argumentIndex]);
                                                    }
                                                    else
                                                    {
                                                        result = "wrong # args: should be \"sql open ?options? connectionString\"";
                                                    }

                                                    code = ReturnCode.Error;
                                                }
                                            }
                                        }
                                        else
                                        {
                                            result = "wrong # args: should be \"sql open ?options? connectionString\"";
                                            code = ReturnCode.Error;
                                        }
                                        break;
                                    }
                                case "transaction":
                                    {
                                        if (arguments.Count >= 4)
                                        {
                                            OptionDictionary options = new OptionDictionary(
                                                new Option[] {
                                                new Option(typeof(IsolationLevel), OptionFlags.MustHaveEnumValue, Index.Invalid, Index.Invalid, "-isolation", null),
                                                new Option(null, OptionFlags.None, Index.Invalid, Index.Invalid, Option.EndOfOptions, null)
                                            });

                                            int argumentIndex = Index.Invalid;

                                            code = interpreter.GetOptions(options, arguments, 0, 2, Index.Invalid, false, ref argumentIndex, ref result);

                                            if (code == ReturnCode.Ok)
                                            {
                                                if ((argumentIndex != Index.Invalid) &&
                                                    ((argumentIndex + 2) == arguments.Count))
                                                {
                                                    Variant value = null;
                                                    IsolationLevel isolationLevel = IsolationLevel.Unspecified; /* NOTE: Yes, this default is OK, per MSDN. */

                                                    if (options.IsPresent("-isolation", ref value))
                                                        isolationLevel = (IsolationLevel)value.Value;

                                                    switch (arguments[argumentIndex])
                                                    {
                                                        case "begin":
                                                            {
                                                                IDbConnection connection = null;

                                                                code = interpreter.GetDbConnection(arguments[argumentIndex + 1], ref connection, ref result);

                                                                if (code == ReturnCode.Ok)
                                                                {
                                                                    //
                                                                    // NOTE: We are going to modify the interpreter state, make
                                                                    //       sure it is not set to read-only.  Technically, this
                                                                    //       modifies the interpreter state directly (via the
                                                                    //       transactions dictionary); however, we may need to
                                                                    //       relax or remove this read-only restriction in the
                                                                    //       future.
                                                                    //
                                                                    if (!interpreter.ReadOnly)
                                                                    {
                                                                        if (interpreter.HasDbTransactions(ref result))
                                                                        {
                                                                            try
                                                                            {
                                                                                IDbTransaction transaction =
                                                                                    connection.BeginTransaction(isolationLevel);

                                                                                result = FormatOps.DatabaseObjectName(transaction,
                                                                                    "Transaction", interpreter.NextId());

                                                                                lock (interpreter.SyncRoot)
                                                                                {
                                                                                    interpreter.transactions.Add(result, transaction);
                                                                                }

#if NOTIFY
                                                                                /* IGNORED */
                                                                                interpreter.CheckNotifications(
                                                                                    NotifyType.Transaction, NotifyFlags.Added,
                                                                                    transaction, interpreter, null, null, null,
                                                                                    ref result);
#endif
                                                                            }
                                                                            catch (Exception e)
                                                                            {
                                                                                Engine.SetExceptionErrorCode(interpreter, e);

                                                                                result = e;
                                                                                code = ReturnCode.Error;
                                                                            }
                                                                        }
                                                                        else
                                                                        {
                                                                            code = ReturnCode.Error;
                                                                        }
                                                                    }
                                                                    else
                                                                    {
                                                                        result = "interpreter is read-only";
                                                                        code = ReturnCode.Error;
                                                                    }
                                                                }
                                                                break;
                                                            }
                                                        case "commit":
                                                            {
                                                                IDbTransaction transaction = null;

                                                                code = interpreter.GetDbTransaction(arguments[argumentIndex + 1], ref transaction, ref result);

                                                                if (code == ReturnCode.Ok)
                                                                {
                                                                    //
                                                                    // NOTE: We are going to modify the interpreter state, make
                                                                    //       sure it is not set to read-only.  Technically, this
                                                                    //       modifies the interpreter state directly (via the
                                                                    //       transactions dictionary); however, we may need to
                                                                    //       relax or remove this read-only restriction in the
                                                                    //       future.
                                                                    //
                                                                    if (!interpreter.ReadOnly)
                                                                    {
                                                                        if (interpreter.HasDbTransactions(ref result))
                                                                        {
                                                                            try
                                                                            {
                                                                                transaction.Commit();

                                                                                lock (interpreter.SyncRoot)
                                                                                {
                                                                                    interpreter.transactions.Remove(arguments[argumentIndex + 1]);
                                                                                }

#if NOTIFY
                                                                                /* IGNORED */
                                                                                interpreter.CheckNotifications(
                                                                                    NotifyType.Transaction, NotifyFlags.Removed,
                                                                                    transaction, interpreter, null, null, null,
                                                                                    ref result);
#endif

                                                                                transaction = null;
                                                                            }
                                                                            catch (Exception e)
                                                                            {
                                                                                Engine.SetExceptionErrorCode(interpreter, e);

                                                                                result = e;
                                                                                code = ReturnCode.Error;
                                                                            }
                                                                        }
                                                                        else
                                                                        {
                                                                            code = ReturnCode.Error;
                                                                        }
                                                                    }
                                                                    else
                                                                    {
                                                                        result = "interpreter is read-only";
                                                                        code = ReturnCode.Error;
                                                                    }
                                                                }
                                                                break;
                                                            }
                                                        case "rollback":
                                                            {
                                                                IDbTransaction transaction = null;

                                                                code = interpreter.GetDbTransaction(arguments[argumentIndex + 1], ref transaction, ref result);

                                                                if (code == ReturnCode.Ok)
                                                                {
                                                                    //
                                                                    // NOTE: We are going to modify the interpreter state, make
                                                                    //       sure it is not set to read-only.  Technically, this
                                                                    //       modifies the interpreter state directly (via the
                                                                    //       transactions dictionary); however, we may need to
                                                                    //       relax or remove this read-only restriction in the
                                                                    //       future.
                                                                    //
                                                                    if (!interpreter.ReadOnly)
                                                                    {
                                                                        if (interpreter.HasDbTransactions(ref result))
                                                                        {
                                                                            try
                                                                            {
                                                                                transaction.Rollback();

                                                                                lock (interpreter.SyncRoot)
                                                                                {
                                                                                    interpreter.transactions.Remove(arguments[argumentIndex + 1]);
                                                                                }

#if NOTIFY
                                                                                /* IGNORED */
                                                                                interpreter.CheckNotifications(
                                                                                    NotifyType.Transaction, NotifyFlags.Removed,
                                                                                    transaction, interpreter, null, null, null,
                                                                                    ref result);
#endif

                                                                                transaction = null;
                                                                            }
                                                                            catch (Exception e)
                                                                            {
                                                                                Engine.SetExceptionErrorCode(interpreter, e);

                                                                                result = e;
                                                                                code = ReturnCode.Error;
                                                                            }
                                                                        }
                                                                        else
                                                                        {
                                                                            code = ReturnCode.Error;
                                                                        }
                                                                    }
                                                                    else
                                                                    {
                                                                        result = "interpreter is read-only";
                                                                        code = ReturnCode.Error;
                                                                    }
                                                                }
                                                                break;
                                                            }
                                                        default:
                                                            {
                                                                result = String.Format(
                                                                    "unsupported database transaction action \"{0}\"",
                                                                    arguments[argumentIndex]);

                                                                code = ReturnCode.Error;
                                                                break;
                                                            }
                                                    }
                                                }
                                                else
                                                {
                                                    if ((argumentIndex != Index.Invalid) &&
                                                        Option.LooksLikeOption(arguments[argumentIndex]))
                                                    {
                                                        result = OptionDictionary.BadOption(options, arguments[argumentIndex]);
                                                    }
                                                    else
                                                    {
                                                        result = "wrong # args: should be \"sql transaction ?options? action object\"";
                                                    }

                                                    code = ReturnCode.Error;
                                                }
                                            }
                                        }
                                        else
                                        {
                                            result = "wrong # args: should be \"sql transaction ?options? action object\"";
                                            code = ReturnCode.Error;
                                        }
                                        break;
                                    }
                                default:
                                    {
                                        result = RuntimeOps.BadSubCommand(null, null, subCommand, this, null, null);
                                        code = ReturnCode.Error;
                                        break;
                                    }
                            }
                        }
                    }
                    else
                    {
                        result = "wrong # args: should be \"sql option ?arg ...?\"";
                        code = ReturnCode.Error;
                    }
                }
                else
                {
                    result = "invalid argument list";
                    code = ReturnCode.Error;
                }
            }
            else
            {
                result = "invalid interpreter";
                code = ReturnCode.Error;
            }

            return code;
        }
        #endregion
    }
}
