/*
 * CommonOps.cs --
 *
 * Copyright (c) 2007-2012 by Joe Mistachkin.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution of
 * this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 * RCS: @(#) $Id: $
 */

using System;
using System.Text.RegularExpressions;
using Eagle._Attributes;
using Eagle._Components.Public;
using Eagle._Containers.Private;

namespace Eagle._Components.Private
{
    [ObjectId("c385e1b9-95b0-4cd5-b0d9-a5fe582d7162")]
    internal static class CommonOps
    {
        [ObjectId("e9622641-301b-4208-a5cc-3801edf4854e")]
        internal static class Runtime
        {
            private static readonly string MonoRuntimeType = "Mono.Runtime"; // NOTE: Only present in Mono.
            private static readonly string MonoDisplayNameMember = "GetDisplayName";

            //
            // NOTE: Create a list of regular expression patterns to check the
            //       Mono runtime version against.
            //
            private static readonly RegExList MonoVersionRegExList = new RegExList(new Regex[] {
                new Regex(" (\\d+(?:\\.\\d+)+)$", /* NOTE: Pre-2.6.0? */
                    RegexOptions.IgnoreCase | RegexOptions.CultureInvariant),
                new Regex("^(\\d+(?:\\.\\d+)+) ", /* NOTE: Post-2.6.0? */
                    RegexOptions.IgnoreCase | RegexOptions.CultureInvariant)
            });

            ///////////////////////////////////////////////////////////////////////////////////////////////

            private static readonly string MonoRuntimeName = "Mono";
            private static readonly string MicrosoftRuntimeName = "Microsoft.NET";

            ///////////////////////////////////////////////////////////////////////////////////////////////

            private static bool? isMono = null;

            ///////////////////////////////////////////////////////////////////////////////////////////////

            private static Version FrameworkVersion = Environment.Version;
            private static OperatingSystem OperatingSystem = Environment.OSVersion;

            ///////////////////////////////////////////////////////////////////////////////////////////////

            public static bool IsMono()
            {
                try
                {
                    if (isMono == null)
                        isMono = (Type.GetType(MonoRuntimeType) != null);

                    return (bool)isMono;
                }
                catch
                {
                    // do nothing.
                }

                return false;
            }

            ///////////////////////////////////////////////////////////////////////////////////////////////

            private static Version GetMonoVersion()
            {
                try
                {
                    Type type = Type.GetType(MonoRuntimeType);

                    if (type != null)
                    {
                        string displayName = type.InvokeMember(
                            MonoDisplayNameMember,
                            MarshalOps.PrivateStaticMethodBindingFlags,
                            null, null, null) as string;

                        if (!String.IsNullOrEmpty(displayName))
                        {
                            if (MonoVersionRegExList != null)
                            {
                                foreach (Regex regEx in MonoVersionRegExList)
                                {
                                    Match match = regEx.Match(displayName);

                                    if ((match != null) && match.Success)
                                        return new Version(match.Value);
                                }
                            }
                        }
                    }
                }
                catch
                {
                    // do nothing.
                }

                return null;
            }

            ///////////////////////////////////////////////////////////////////////////////////////////////

            public static string GetRuntimeName()
            {
                return IsMono() ? MonoRuntimeName : MicrosoftRuntimeName;
            }

            ///////////////////////////////////////////////////////////////////////////////////////////////

            public static Version GetRuntimeVersion()
            {
                return IsMono() ? GetMonoVersion() : GetFrameworkVersion();
            }

            ///////////////////////////////////////////////////////////////////////////////////////////////

            public static string GetRuntimeNameAndVersion()
            {
                return FormatOps.NameAndVersion(GetRuntimeName(), GetRuntimeVersion());
            }

            ///////////////////////////////////////////////////////////////////////////////////////////////

            public static Version GetFrameworkVersion()
            {
                return FrameworkVersion;
            }

            ///////////////////////////////////////////////////////////////////////////////////////////////

            public static bool IsFramework2() // NOTE: Be sure to use !IsMono also.
            {
                Version version = GetFrameworkVersion();
                return (version != null) && (version.Major == 2);
            }

            ///////////////////////////////////////////////////////////////////////////////////////////////

            public static bool IsFramework4() // NOTE: Be sure to use !IsMono also.
            {
                Version version = GetFrameworkVersion();
                return (version != null) && (version.Major == 4);
            }

            ///////////////////////////////////////////////////////////////////////////////////////////////

            public static OperatingSystem GetOperatingSystem()
            {
                return OperatingSystem;
            }

            ///////////////////////////////////////////////////////////////////////////////////////////////

            public static PlatformID GetOperatingSystemPlatform()
            {
                OperatingSystem operatingSystem = GetOperatingSystem();

                return (operatingSystem != null) ?
                    operatingSystem.Platform : (PlatformID)(int)OperatingSystemId.Unknown;
            }

            ///////////////////////////////////////////////////////////////////////////////////////////////

            public static Version GetOperatingSystemVersion()
            {
                OperatingSystem operatingSystem = GetOperatingSystem();

                return (operatingSystem != null) ? operatingSystem.Version : null;
            }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public static int CombineHashCodes(
            int X,
            int Y
            )
        {
            byte[] bytes = new byte[sizeof(int) * 2];

            Array.Copy(BitConverter.GetBytes(X), 0, bytes, 0, sizeof(int));
            Array.Copy(BitConverter.GetBytes(Y), 0, bytes, sizeof(int), sizeof(int));

            return ConversionOps.ToInt(MathOps.HashFnv1UInt(bytes, true));
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public static int CombineHashCodes(
            int X,
            int Y,
            int Z
            )
        {
            byte[] bytes = new byte[sizeof(int) * 3];

            Array.Copy(BitConverter.GetBytes(X), 0, bytes, 0, sizeof(int));
            Array.Copy(BitConverter.GetBytes(Y), 0, bytes, sizeof(int), sizeof(int));
            Array.Copy(BitConverter.GetBytes(Z), 0, bytes, sizeof(int) * 2, sizeof(int));

            return ConversionOps.ToInt(MathOps.HashFnv1UInt(bytes, true));
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public static bool ForCondition(
            bool increment,
            int index,
            int lowerBound,
            int upperBound
            )
        {
            if (increment)
                return (index <= upperBound);
            else
                return (index >= lowerBound);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public static void ForLoop(
            bool increment,
            ref int index
            )
        {
            if (increment)
                index++;
            else
                index--;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public static string GetEnvironmentVariable(
            string variable
            )
        {
            return Environment.GetEnvironmentVariable(variable);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public static void SetEnvironmentVariable(
            string variable,
            string value
            )
        {
            Environment.SetEnvironmentVariable(variable, value);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public static string ExpandEnvironmentVariables(
            string name
            )
        {
            return Environment.ExpandEnvironmentVariables(name);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public static bool DoesEnvironmentVariableExist(
            string name
            )
        {
            string value = null;

            return DoesEnvironmentVariableExist(name, ref value);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public static bool DoesEnvironmentVariableExist(
            string name,
            ref string value
            )
        {
            value = GetEnvironmentVariable(name);
            return value != null;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private static bool ShouldGC()
        {
            //
            // NOTE: If this environment variable is set, the Eagle library
            //       will never manually call into the GC to have it collect
            //       garbage; otherwise, manual calls into the GC will be
            //       enabled at certain strategic points in the code where
            //       it makes sense.
            //
            if (GetEnvironmentVariable(EnvVars.NeverGC) != null)
            {
                return false;
            }

            return true;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private static bool ShouldWaitForGC()
        {
            //
            // NOTE: If this environment variable is set, always wait for
            //       the GC to finish the pending finalizers; otherwise,
            //       we will only wait if this is the default application
            //       domain to prevent a subtle deadlock that can seemingly
            //       occur in applications that contain a user-interface
            //       that may be running in an isolated application domain
            //       (see below).  Otherwise, if the "opposite" environment
            //       variable is set, never wait for the GC to finish the
            //       pending finalizers.
            //
            if (GetEnvironmentVariable(EnvVars.AlwaysWaitForGC) != null)
            {
                return true;
            }
            else if (GetEnvironmentVariable(EnvVars.NeverWaitForGC) != null)
            {
                return false;
            }
            else
            {
                //
                // BUGBUG: Only wait for pending finalizers in the default
                //         application domain (due to potential deadlocks?).
                //         This seems to be related to the cross-AppDomain
                //         marshalling in .NET wanting to obtain a lock on
                //         the GC from two threads at the same time, which
                //         results in a deadlock.  This issue was observed
                //         in a WPF application loaded into an isolated
                //         application domain; therefore, this issue may be
                //         limited to applications that contain some kind
                //         of user-interface thread.
                //
                if (RuntimeOps.IsDefaultAppDomain())
                    return true;
            }

            return false;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private static void CollectGarbage(
            int generation,
            GCCollectionMode collectionMode
            ) /* throw */
        {
            if (generation == -1)
                GC.Collect();
            else
                GC.Collect(generation, collectionMode);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public static void CollectGarbage(
            bool? force,
            bool? wait
            ) /* throw */
        {
            CollectGarbage(-1, GCCollectionMode.Default, force, wait);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private static void CollectGarbage(
            int generation,
            GCCollectionMode collectionMode,
            bool? force,
            bool? wait
            ) /* throw */
        {
            //
            // NOTE: If the force parameter is null, that means we should
            //       attempt to automatically detect whether or not we
            //       should actually collect any garbage.
            //
            bool reallyCollect = (force == null) ? ShouldGC() : (bool)force;

            if (!reallyCollect)
                return;

            //
            // NOTE: If the wait parameter is null, that means we should
            //       attempt to automatically detect whether or not we
            //       should wait for the pending finalizers to finish.
            //
            bool reallyWait = (wait == null) ? ShouldWaitForGC() : (bool)wait;

            CollectGarbage(generation, collectionMode);

            if (reallyWait)
            {
                GC.WaitForPendingFinalizers();
                CollectGarbage(generation, collectionMode);
            }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

#if SHELL
        public static void GetTotalMemory(
            bool collect,
            ref long beforeBytes,
            ref long afterBytes
            )
        {
            beforeBytes = GC.GetTotalMemory(false);

            if (collect)
                afterBytes = GC.GetTotalMemory(true);
        }
#endif
    }
}
