/*
 * ConsoleOps.cs --
 *
 * Copyright (c) 2007-2012 by Joe Mistachkin.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution of
 * this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 * RCS: @(#) $Id: $
 */

#if !NATIVE || !WINDOWS
#error "This file cannot be compiled or used properly with native Windows code disabled."
#endif

using System;
using System.Runtime.InteropServices;
using System.Security;

#if !NET_40
using System.Security.Permissions;
#endif

using Eagle._Attributes;
using Eagle._Components.Public;
using Eagle._Constants;

namespace Eagle._Components.Private
{
#if NET_40
    [SecurityCritical()]
#else
    [SecurityPermission(SecurityAction.LinkDemand, UnmanagedCode = true)]
#endif
    [ObjectId("7b199e66-8290-4cdc-8312-d02de62683c5")]
    internal static class ConsoleOps
    {
        #region Private Constants
        private static readonly bool DefaultNative = true; /* CommonOps.Runtime.IsMono(); */
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////
        // Required Native APIs used via P/Invoke
        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region Unsafe Native Methods Class
        [SuppressUnmanagedCodeSecurity()]
        [ObjectId("6c20c88a-dd55-4e35-a5a5-ec288041c8f4")]
        internal static class UnsafeNativeMethods
        {
            //
            // NOTE: Console input modes.
            //
            internal const uint ENABLE_MOUSE_INPUT = 0x10;

            //
            // NOTE: Console output modes.
            //
            //internal const uint ENABLE_PROCESSED_OUTPUT = 0x01;

            //
            // NOTE: Win32 error numbers.
            //
            internal const int NO_ERROR = 0;
            internal const int ERROR_INVALID_HANDLE = 6;

            //
            // NOTE: Values returned by GetFileType.
            //
            internal const uint FILE_TYPE_UNKNOWN = 0x0;
            internal const uint FILE_TYPE_DISK = 0x1;
            internal const uint FILE_TYPE_CHAR = 0x2;
            internal const uint FILE_TYPE_PIPE = 0x3;
            internal const uint FILE_TYPE_REMOTE = 0x8000;

            //
            // NOTE: Console handles.
            //
            internal const int STD_INPUT_HANDLE = -10;
            internal const int STD_OUTPUT_HANDLE = -11;
            internal const int STD_ERROR_HANDLE = -12;

            #region Dead Code
#if DEAD_CODE
            //
            // NOTE: Special process id.
            //
            internal const int ATTACH_PARENT_PROCESS = -1;
#endif
            #endregion

            [DllImport(DllName.Kernel32, CallingConvention = CallingConvention.Winapi, SetLastError = true)]
            internal static extern uint GetFileType(IntPtr handle);

            [DllImport(DllName.Kernel32, CallingConvention = CallingConvention.Winapi, SetLastError = true)]
            [return: MarshalAs(UnmanagedType.Bool)]
            internal static extern bool GetConsoleMode(IntPtr handle, ref uint mode);

            //[DllImport(DllName.Kernel32, CallingConvention = CallingConvention.Winapi, SetLastError = true)]
            //internal static extern IntPtr GetConsoleInputWaitHandle(); /* UNDOCUMENTED */

            [DllImport(DllName.Kernel32, CallingConvention = CallingConvention.Winapi, SetLastError = true)]
            [return: MarshalAs(UnmanagedType.Bool)]
            internal static extern bool SetConsoleMode(IntPtr handle, uint mode);

            [DllImport(DllName.Kernel32, CallingConvention = CallingConvention.Winapi, SetLastError = true)]
            [return: MarshalAs(UnmanagedType.Bool)]
            internal static extern bool GenerateConsoleCtrlEvent(ControlEvent controlEvent, uint processGroupId);

            [DllImport(DllName.Kernel32, CallingConvention = CallingConvention.Winapi, SetLastError = true)]
            internal static extern IntPtr GetStdHandle(int nStdHandle);

            #region Dead Code
#if DEAD_CODE
            [DllImport(DllName.Kernel32, CallingConvention = CallingConvention.Winapi, SetLastError = true)]
            [return: MarshalAs(UnmanagedType.Bool)]
            private static extern bool SetStdHandle(int nStdHandle, IntPtr handle);
#endif
            #endregion

            [DllImport(DllName.Kernel32, CallingConvention = CallingConvention.Winapi)]
            internal static extern IntPtr GetConsoleWindow();

            [DllImport(DllName.Kernel32, CallingConvention = CallingConvention.Winapi, SetLastError = true)]
            [return: MarshalAs(UnmanagedType.Bool)]
            internal static extern bool FlushConsoleInputBuffer(IntPtr handle);

            #region Dead Code
#if DEAD_CODE
            [DllImport(DllName.Kernel32, CallingConvention = CallingConvention.Winapi, SetLastError = true)]
            [return: MarshalAs(UnmanagedType.Bool)]
            internal static extern bool AttachConsole(int processId);
#endif
            #endregion

            [DllImport(DllName.Kernel32, CallingConvention = CallingConvention.Winapi, SetLastError = true)]
            [return: MarshalAs(UnmanagedType.Bool)]
            internal static extern bool AllocConsole();

            [DllImport(DllName.Kernel32, CallingConvention = CallingConvention.Winapi, SetLastError = true)]
            [return: MarshalAs(UnmanagedType.Bool)]
            internal static extern bool FreeConsole();
        }
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public static IntPtr GetHandle(
            ChannelType channelType,
            ref Result error
            )
        {
            return GetHandle(channelType, DefaultNative, ref error);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private static IntPtr GetHandle(
            ChannelType channelType,
            bool native,
            ref Result error
            )
        {
            switch (channelType)
            {
                case ChannelType.Input:
                    return GetInputHandle(native, ref error);
                case ChannelType.Output:
                    return GetOutputHandle(native, ref error);
                case ChannelType.Error:
                    return GetErrorHandle(native, ref error);
                default:
                    error = "unsupported console channel";
                    return IntPtr.Zero;
            }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private static IntPtr GetInputHandle(
            bool native,
            ref Result error
            )
        {
            IntPtr handle = IntPtr.Zero;

            try
            {
                if (native)
                {
                    handle = UnsafeNativeMethods.GetStdHandle(
                        UnsafeNativeMethods.STD_INPUT_HANDLE);

                    if (!NativeOps.IsValidHandle(handle))
                        error = NativeOps.GetErrorMessage(
                            Marshal.GetLastWin32Error());
                }
                else
                {
#if CONSOLE
                    handle = (IntPtr)typeof(Console).InvokeMember("ConsoleInputHandle",
                        MarshalOps.PrivateStaticGetPropertyBindingFlags, null, null, null);

                    if (!NativeOps.IsValidHandle(handle))
                        error = "invalid console input handle";
#else
                    error = "not implemented";
#endif
                }
            }
            catch (Exception e)
            {
                error = e;
            }

            return handle;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private static IntPtr GetOutputHandle(
            bool native,
            ref Result error
            )
        {
            IntPtr handle = IntPtr.Zero;

            try
            {
                if (native)
                {
                    handle = UnsafeNativeMethods.GetStdHandle(
                        UnsafeNativeMethods.STD_OUTPUT_HANDLE);

                    if (!NativeOps.IsValidHandle(handle))
                        error = NativeOps.GetErrorMessage(
                            Marshal.GetLastWin32Error());
                }
                else
                {
#if CONSOLE
                    handle = (IntPtr)typeof(Console).InvokeMember("ConsoleOutputHandle",
                        MarshalOps.PrivateStaticGetPropertyBindingFlags, null, null, null);

                    if (!NativeOps.IsValidHandle(handle))
                        error = "invalid console output handle";
#else
                    error = "not implemented";
#endif
                }
            }
            catch (Exception e)
            {
                error = e;
            }

            return handle;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private static IntPtr GetErrorHandle(
            bool native,
            ref Result error
            )
        {
            IntPtr handle = IntPtr.Zero;

            try
            {
                if (native)
                {
                    //
                    // NOTE: This is always done natively.  The System.Console class does not
                    //       keep track of the standard error channel.
                    //
                    handle = UnsafeNativeMethods.GetStdHandle(
                        UnsafeNativeMethods.STD_ERROR_HANDLE);

                    if (!NativeOps.IsValidHandle(handle))
                        error = NativeOps.GetErrorMessage(
                            Marshal.GetLastWin32Error());
                }
                else
                {
                    error = "not implemented";
                }
            }
            catch (Exception e)
            {
                error = e;
            }

            return handle;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public static ReturnCode IsHandleRedirected(
            IntPtr handle,
            ref bool redirected,
            ref Result error
            )
        {
            if (NativeOps.IsValidHandle(handle))
            {
                try
                {
                    uint type = UnsafeNativeMethods.GetFileType(handle);

                    if ((type != UnsafeNativeMethods.FILE_TYPE_UNKNOWN) ||
                        (Marshal.GetLastWin32Error() == UnsafeNativeMethods.NO_ERROR))
                    {
                        type &= ~UnsafeNativeMethods.FILE_TYPE_REMOTE;

                        if (type == UnsafeNativeMethods.FILE_TYPE_CHAR)
                        {
                            uint mode = 0;

                            if (UnsafeNativeMethods.GetConsoleMode(handle, ref mode))
                            {
                                //
                                // NOTE: We do not care about the mode, this is a
                                //       console simply because GetConsoleMode
                                //       succeeded.
                                //
                                redirected = false;
                            }
                            else if (Marshal.GetLastWin32Error() ==
                                    UnsafeNativeMethods.ERROR_INVALID_HANDLE)
                            {
                                //
                                // NOTE: The handle appears to be valid (see above)
                                //       and does not represent a console; therefore,
                                //       it has been redirected to something that is
                                //       not a console.
                                //
                                redirected = true;
                            }
                            else
                            {
                                //
                                // NOTE: The handle appears to be valid (see above)
                                //       and must be a console because GetConsoleMode
                                //       did not set the error to ERROR_INVALID_HANDLE.
                                //
                                redirected = false;
                            }
                        }
                        else
                        {
                            //
                            // NOTE: The handle appears to be valid (see above); It
                            //       cannot be a console because it is not a character
                            //       device; therefore, it must have been redirected.
                            //
                            redirected = true;
                        }
                    }
                    else
                    {
                        //
                        // NOTE: The handle appears to be valid; however, we
                        //       cannot determine the file type.  We must assume
                        //       that it has not been redirected.
                        //
                        redirected = false;
                    }

                    return ReturnCode.Ok;
                }
                catch (Exception e)
                {
                    error = e;
                }
            }
            else
            {
                error = "invalid handle";
            }

            return ReturnCode.Error;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public static ReturnCode FlushInputBuffer(
            ref Result error
            )
        {
            try
            {
                IntPtr handle = GetHandle(
                    ChannelType.Input, DefaultNative, ref error);

                if (NativeOps.IsValidHandle(handle))
                {
                    if (UnsafeNativeMethods.FlushConsoleInputBuffer(handle))
                        return ReturnCode.Ok;

                    error = NativeOps.GetErrorMessage(
                        Marshal.GetLastWin32Error());
                }
            }
            catch (Exception e)
            {
                error = e;
            }

            return ReturnCode.Error;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region Dead Code
#if DEAD_CODE
        public static ReturnCode Attach(
            ref Result error
            )
        {
            try
            {
                if (UnsafeNativeMethods.AttachConsole(
                        UnsafeNativeMethods.ATTACH_PARENT_PROCESS))
                    return ReturnCode.Ok;
                else
                    error = NativeOps.GetErrorMessage(
                        Marshal.GetLastWin32Error());
            }
            catch (Exception e)
            {
                error = e;
            }

            return ReturnCode.Error;
        }
#endif
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public static ReturnCode Open(
            ref Result error
            )
        {
            try
            {
                if (UnsafeNativeMethods.AllocConsole())
                    return ReturnCode.Ok;
                else
                    error = NativeOps.GetErrorMessage(
                        Marshal.GetLastWin32Error());
            }
            catch (Exception e)
            {
                error = e;
            }

            return ReturnCode.Error;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public static ReturnCode Close(
            ref Result error
            )
        {
            try
            {
                if (UnsafeNativeMethods.FreeConsole())
                    return ReturnCode.Ok;
                else
                    error = NativeOps.GetErrorMessage(
                        Marshal.GetLastWin32Error());
            }
            catch (Exception e)
            {
                error = e;
            }

            return ReturnCode.Error;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public static ReturnCode GetMode(
            ChannelType channelType,
            ref uint mode,
            ref Result error
            )
        {
            try
            {
                IntPtr handle = GetHandle(channelType, DefaultNative, ref error);

                if (NativeOps.IsValidHandle(handle))
                {
                    if (UnsafeNativeMethods.GetConsoleMode(handle, ref mode))
                        return ReturnCode.Ok;

                    error = NativeOps.GetErrorMessage(
                        Marshal.GetLastWin32Error());
                }
            }
            catch (Exception e)
            {
                error = e;
            }

            return ReturnCode.Error;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public static ReturnCode SetMode(
            ChannelType channelType,
            uint mode,
            ref Result error
            )
        {
            try
            {
                IntPtr handle = GetHandle(channelType, DefaultNative, ref error);

                if (NativeOps.IsValidHandle(handle))
                {
                    if (UnsafeNativeMethods.SetConsoleMode(handle, mode))
                        return ReturnCode.Ok;

                    error = NativeOps.GetErrorMessage(
                        Marshal.GetLastWin32Error());
                }
            }
            catch (Exception e)
            {
                error = e;
            }

            return ReturnCode.Error;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public static ReturnCode ChangeMode(
            ChannelType channelType,
            bool enable,
            uint mode
            )
        {
            Result error = null;

            return ChangeMode(channelType, enable, mode, ref error);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private static ReturnCode ChangeMode(
            ChannelType channelType,
            bool enable,
            uint mode,
            ref Result error
            )
        {
            try
            {
                IntPtr handle = GetHandle(channelType, DefaultNative, ref error);

                if (NativeOps.IsValidHandle(handle))
                {
                    uint currentMode = 0;

                    if (UnsafeNativeMethods.GetConsoleMode(handle, ref currentMode))
                    {
                        if (enable)
                            currentMode |= mode;  /* NOTE: Add mode(s). */
                        else
                            currentMode &= ~mode; /* NOTE: Remove mode(s). */

                        if (UnsafeNativeMethods.SetConsoleMode(handle, currentMode))
                            return ReturnCode.Ok;
                    }

                    error = NativeOps.GetErrorMessage(
                        Marshal.GetLastWin32Error());
                }
            }
            catch (Exception e)
            {
                error = e;
            }

            return ReturnCode.Error;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public static ReturnCode SendControlEvent(
            ControlEvent @event,
            ref Result error
            )
        {
            try
            {
                if (UnsafeNativeMethods.GenerateConsoleCtrlEvent(@event, 0))
                    return ReturnCode.Ok;
                else
                    error = NativeOps.GetErrorMessage(
                        Marshal.GetLastWin32Error());
            }
            catch (Exception e)
            {
                error = e;
            }

            return ReturnCode.Error;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public static ReturnCode CloseStandardInput(
            ref Result error
            )
        {
            try
            {
                IntPtr handle = GetOutputHandle(DefaultNative, ref error);

                if (NativeOps.IsValidHandle(handle))
                {
                    if (NativeOps.UnsafeNativeMethods.CloseHandle(handle))
                        return ReturnCode.Ok;

                    error = NativeOps.GetErrorMessage(
                        Marshal.GetLastWin32Error());
                }
            }
            catch (Exception e)
            {
                error = e;
            }

            return ReturnCode.Error;
        }
    }
}

