/*
 * ContextManager.cs --
 *
 * Copyright (c) 2007-2012 by Joe Mistachkin.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution of
 * this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 * RCS: @(#) $Id: $
 */

using System;
using System.Threading;
using Eagle._Attributes;
using Eagle._Components.Public;
using Eagle._Containers.Private;
using Eagle._Containers.Public;
using Eagle._Interfaces.Private;
using Eagle._Interfaces.Public;

namespace Eagle._Components.Private
{
    [ObjectId("53f5f2e0-ea9c-46a8-b584-779ef217beb5")]
    internal sealed class ContextManager :
            IContextManager, IHaveInterpreter, IDisposable
    {
        #region Private Static Data
        private static object syncRoot = new object();

        private static LocalDataStoreSlot engineSlot;
        private static LocalDataStoreSlot interactiveSlot;
        private static LocalDataStoreSlot testSlot;
        private static LocalDataStoreSlot variableSlot;
        #endregion

        ///////////////////////////////////////////////////////////////////////

        #region Private Data
        private IEngineContext previousEngineContext;
        private IInteractiveContext previousInteractiveContext;
        private ITestContext previousTestContext;
        private IVariableContext previousVariableContext;
        #endregion

        ///////////////////////////////////////////////////////////////////////

        #region Public Static Methods
        public static void Initialize()
        {
            //
            // NOTE: These MUST to be done prior to evaluating any scripts or
            //       call frame handling (and a bunch of other stuff) will not
            //       work properly.
            //
            lock (syncRoot)
            {
                if (engineSlot == null)
                    engineSlot = Thread.AllocateDataSlot();

                if (interactiveSlot == null)
                    interactiveSlot = Thread.AllocateDataSlot();

                if (testSlot == null)
                    testSlot = Thread.AllocateDataSlot();

                if (variableSlot == null)
                    variableSlot = Thread.AllocateDataSlot();
            }
        }
        #endregion

        ///////////////////////////////////////////////////////////////////////

        #region Public Constructors
        public ContextManager(
            Interpreter interpreter
            )
        {
            this.interpreter = interpreter;
        }
        #endregion

        ///////////////////////////////////////////////////////////////////////

        #region Private Static Methods
        private static int GetCurrentThreadId()
        {
            return GlobalState.GetCurrentThreadId();
        }

        ///////////////////////////////////////////////////////////////////////

        private static bool CanUseThreadContext(
            IThreadContext threadContext
            )
        {
            return (threadContext != null) &&
                (threadContext.ThreadId == GetCurrentThreadId());
        }
        #endregion

        ///////////////////////////////////////////////////////////////////////

        #region IGetInterpreter / ISetInterpreter Members
        private Interpreter interpreter;
        public Interpreter Interpreter
        {
            get { CheckDisposed(); return interpreter; }
            set { CheckDisposed(); interpreter = value; }
        }
        #endregion

        ///////////////////////////////////////////////////////////////////////

        #region IContextManager Members
        public bool FreeEngineContext()
        {
            CheckDisposed();

            Result error = null;

            return FreeEngineContext(ref error);
        }

        ///////////////////////////////////////////////////////////////////////

        public bool FreeEngineContext(
            ref Result error
            )
        {
            CheckDisposed();

            if (interpreter != null)
            {
                try
                {
                    if (engineSlot != null)
                    {
                        //
                        // NOTE: Try to obtain the per-interpreter dictionary
                        //       of contexts for this thread.
                        //
                        InterpreterEngineContextDictionary contexts =
                            Thread.GetData(engineSlot) as
                            InterpreterEngineContextDictionary; /* throw */

                        if (contexts != null)
                        {
                            //
                            // NOTE: Does the requested interpreter have an
                            //       entry in the dictionary?
                            //
                            if (contexts.ContainsKey(interpreter))
                                contexts.Remove(interpreter);

                            //
                            // NOTE: If there are no more contexts present,
                            //       free the collection and the thread data
                            //       slot.
                            //
                            if (contexts.Count == 0)
                            {
                                //
                                // NOTE: Release local reference now.
                                //
                                contexts = null;

                                //
                                // NOTE: Clear it in the per-thread data.
                                //
                                Thread.SetData(engineSlot, contexts);
                            }
                        }

                        //
                        // NOTE: Invalidate the cached context for this
                        //       interpreter.
                        //
                        previousEngineContext = null;

                        return true;
                    }
                    else
                    {
                        error = "engine context slot is invalid";
                    }
                }
                catch (Exception e)
                {
                    error = e;
                }
            }
            else
            {
                error = "invalid interpreter";
            }

            return false;
        }

        ///////////////////////////////////////////////////////////////////////

        public IEngineContext GetEngineContext(
            bool create
            )
        {
            CheckDisposed();

            Result error = null;

            return GetEngineContext(create, ref error);
        }

        ///////////////////////////////////////////////////////////////////////

        public IEngineContext GetEngineContext(
            bool create,
            ref Result error
            )
        {
            CheckDisposed();

            IEngineContext result = null;

            if (interpreter != null)
            {
                try
                {
                    result = previousEngineContext;

                    if (CanUseThreadContext(result))
                    {
                        return result;
                    }
                    else if (engineSlot != null)
                    {
                        //
                        // NOTE: Reset the result because it may still contain
                        //       the previous (possibly invalid) context.
                        //
                        result = null;

                        //
                        // NOTE: Try to obtain the per-interpreter dictionary
                        //       of engine contexts for this thread.
                        //
                        InterpreterEngineContextDictionary contexts =
                            Thread.GetData(engineSlot) as
                            InterpreterEngineContextDictionary; /* throw */

                        //
                        // NOTE: Has the dictionary been created yet and/or can
                        //       we create it now?
                        //
                        if (create && (contexts == null))
                        {
                            //
                            // NOTE: Nope, create it now.
                            //
                            contexts = new InterpreterEngineContextDictionary();

                            //
                            // NOTE: Store it in the per-thread data.
                            //
                            Thread.SetData(engineSlot, contexts); /* throw */
                        }

                        //
                        // NOTE: Is the dictionary available now (if not, we
                        //       have been forbidden by the caller from
                        //       automatically creating it).
                        //
                        if (contexts != null)
                        {
                            //
                            // NOTE: Does the requested interpreter have an
                            //       entry in the dictionary?
                            //
                            if (!contexts.ContainsKey(interpreter))
                            {
                                //
                                // NOTE: Now, add it to the dictionary of
                                //       engine contexts.
                                //
                                contexts.Add(interpreter,
                                    new EngineContext(GetCurrentThreadId()));
                            }

                            //
                            // NOTE: Return the engine context for the
                            //       specified interpreter.
                            //
                            result = contexts[interpreter];

                            //
                            // NOTE: Save the resulting context for next time.
                            //
                            previousEngineContext = result;
                        }
                        else
                        {
                            error = "engine contexts not available";
                        }
                    }
                    else
                    {
                        error = "engine context slot is invalid";
                    }
                }
                catch (Exception e)
                {
                    error = e;
                }
            }
            else
            {
                error = "invalid interpreter";
            }

            return result;
        }

        ///////////////////////////////////////////////////////////////////////

        public int GetEngineContextCount()
        {
            CheckDisposed();

            if (engineSlot != null)
            {
                try
                {
                    InterpreterEngineContextDictionary contexts =
                        Thread.GetData(engineSlot) as
                        InterpreterEngineContextDictionary; /* throw */

                    if (contexts != null)
                        return contexts.Count;
                }
                catch
                {
                    // do nothing.
                }
            }

            return 0;
        }

        ///////////////////////////////////////////////////////////////////////

        public bool FreeInteractiveContext()
        {
            CheckDisposed();

            Result error = null;

            return FreeInteractiveContext(ref error);
        }

        ///////////////////////////////////////////////////////////////////////

        public bool FreeInteractiveContext(
            ref Result error
            )
        {
            CheckDisposed();

            if (interpreter != null)
            {
                try
                {
                    if (interactiveSlot != null)
                    {
                        //
                        // NOTE: Try to obtain the per-interpreter dictionary
                        //       of contexts for this thread.
                        //
                        InterpreterInteractiveContextDictionary contexts =
                            Thread.GetData(interactiveSlot) as
                            InterpreterInteractiveContextDictionary; /* throw */

                        if (contexts != null)
                        {
                            //
                            // NOTE: Does the requested interpreter have an
                            //       entry in the dictionary?
                            //
                            if (contexts.ContainsKey(interpreter))
                                contexts.Remove(interpreter);

                            //
                            // NOTE: If there are no more contexts present,
                            //       free the collection and the thread data
                            //       slot.
                            //
                            if (contexts.Count == 0)
                            {
                                //
                                // NOTE: Release local reference now.
                                //
                                contexts = null;

                                //
                                // NOTE: Clear it in the per-thread data.
                                //
                                Thread.SetData(interactiveSlot, contexts);
                            }
                        }

                        //
                        // NOTE: Invalidate the cached context for this
                        //       interpreter.
                        //
                        previousInteractiveContext = null;

                        return true;
                    }
                    else
                    {
                        error = "interactive context slot is invalid";
                    }
                }
                catch (Exception e)
                {
                    error = e;
                }
            }
            else
            {
                error = "invalid interpreter";
            }

            return false;
        }

        ///////////////////////////////////////////////////////////////////////

        public IInteractiveContext GetInteractiveContext(
            bool create
            )
        {
            CheckDisposed();

            Result error = null;

            return GetInteractiveContext(create, ref error);
        }

        ///////////////////////////////////////////////////////////////////////

        public IInteractiveContext GetInteractiveContext(
            bool create,
            ref Result error
            )
        {
            CheckDisposed();

            IInteractiveContext result = null;

            if (interpreter != null)
            {
                try
                {
                    result = previousInteractiveContext;

                    if (CanUseThreadContext(result))
                    {
                        return result;
                    }
                    else if (interactiveSlot != null)
                    {
                        //
                        // NOTE: Reset the result because it may still contain
                        //       the previous (possibly invalid) context.
                        //
                        result = null;

                        //
                        // NOTE: Try to obtain the per-interpreter dictionary
                        //       of interactive contexts for this thread.
                        //
                        InterpreterInteractiveContextDictionary contexts =
                            Thread.GetData(interactiveSlot) as
                            InterpreterInteractiveContextDictionary; /* throw */

                        //
                        // NOTE: Has the dictionary been created yet and/or can
                        //       we create it now?
                        //
                        if (create && (contexts == null))
                        {
                            //
                            // NOTE: Nope, create it now.
                            //
                            contexts = new InterpreterInteractiveContextDictionary();

                            //
                            // NOTE: Store it in the per-thread data.
                            //
                            Thread.SetData(interactiveSlot, contexts); /* throw */
                        }

                        //
                        // NOTE: Is the dictionary available now (if not, we
                        //       have been forbidden by the caller from
                        //       automatically creating it).
                        //
                        if (contexts != null)
                        {
                            //
                            // NOTE: Does the requested interpreter have an
                            //       entry in the dictionary?
                            //
                            if (!contexts.ContainsKey(interpreter))
                            {
                                //
                                // NOTE: Now, add it to the dictionary of
                                //       interactive contexts.
                                //
                                contexts.Add(interpreter,
                                    new InteractiveContext(GetCurrentThreadId()));
                            }

                            //
                            // NOTE: Return the interactive context for the
                            //       specified interpreter.
                            //
                            result = contexts[interpreter];

                            //
                            // NOTE: Save the resulting context for next time.
                            //
                            previousInteractiveContext = result;
                        }
                        else
                        {
                            error = "interactive contexts not available";
                        }
                    }
                    else
                    {
                        error = "interactive context slot is invalid";
                    }
                }
                catch (Exception e)
                {
                    error = e;
                }
            }
            else
            {
                error = "invalid interpreter";
            }

            return result;
        }

        ///////////////////////////////////////////////////////////////////////

        public int GetInteractiveContextCount()
        {
            CheckDisposed();

            if (interactiveSlot != null)
            {
                try
                {
                    InterpreterInteractiveContextDictionary contexts =
                        Thread.GetData(interactiveSlot) as
                        InterpreterInteractiveContextDictionary; /* throw */

                    if (contexts != null)
                        return contexts.Count;
                }
                catch
                {
                    // do nothing.
                }
            }

            return 0;
        }

        ///////////////////////////////////////////////////////////////////////

        public bool FreeTestContext()
        {
            CheckDisposed();

            Result error = null;

            return FreeTestContext(ref error);
        }

        ///////////////////////////////////////////////////////////////////////

        public bool FreeTestContext(
            ref Result error
            )
        {
            CheckDisposed();

            if (interpreter != null)
            {
                try
                {
                    if (testSlot != null)
                    {
                        //
                        // NOTE: Try to obtain the per-interpreter dictionary
                        //       of contexts for this thread.
                        //
                        InterpreterTestContextDictionary contexts =
                            Thread.GetData(testSlot) as
                            InterpreterTestContextDictionary; /* throw */

                        if (contexts != null)
                        {
                            //
                            // NOTE: Does the requested interpreter have an
                            //       entry in the dictionary?
                            //
                            if (contexts.ContainsKey(interpreter))
                                contexts.Remove(interpreter);

                            //
                            // NOTE: If there are no more contexts present,
                            //       free the collection and the thread data
                            //       slot.
                            //
                            if (contexts.Count == 0)
                            {
                                //
                                // NOTE: Release local reference now.
                                //
                                contexts = null;

                                //
                                // NOTE: Clear it in the per-thread data.
                                //
                                Thread.SetData(testSlot, contexts);
                            }
                        }

                        //
                        // NOTE: Invalidate the cached context for this
                        //       interpreter.
                        //
                        previousTestContext = null;

                        return true;
                    }
                    else
                    {
                        error = "test context slot is invalid";
                    }
                }
                catch (Exception e)
                {
                    error = e;
                }
            }
            else
            {
                error = "invalid interpreter";
            }

            return false;
        }

        ///////////////////////////////////////////////////////////////////////

        public ITestContext GetTestContext(
            bool create
            )
        {
            CheckDisposed();

            Result error = null;

            return GetTestContext(create, ref error);
        }

        ///////////////////////////////////////////////////////////////////////

        public ITestContext GetTestContext(
            bool create,
            ref Result error
            )
        {
            CheckDisposed();

            ITestContext result = null;

            if (interpreter != null)
            {
                try
                {
                    result = previousTestContext;

                    if (CanUseThreadContext(result))
                    {
                        return result;
                    }
                    else if (testSlot != null)
                    {
                        //
                        // NOTE: Reset the result because it may still contain
                        //       the previous (possibly invalid) context.
                        //
                        result = null;

                        //
                        // NOTE: Try to obtain the per-interpreter dictionary of
                        //       test contexts for this thread.
                        //
                        InterpreterTestContextDictionary contexts =
                            Thread.GetData(testSlot) as
                            InterpreterTestContextDictionary; /* throw */

                        //
                        // NOTE: Has the dictionary been created yet and/or can
                        //       we create it now?
                        //
                        if (create && (contexts == null))
                        {
                            //
                            // NOTE: Nope, create it now.
                            //
                            contexts = new InterpreterTestContextDictionary();

                            //
                            // NOTE: Store it in the per-thread data.
                            //
                            Thread.SetData(testSlot, contexts); /* throw */
                        }

                        //
                        // NOTE: Is the dictionary available now (if not, we
                        //       have been forbidden by the caller from
                        //       automatically creating it).
                        //
                        if (contexts != null)
                        {
                            //
                            // NOTE: Does the requested interpreter have an
                            //       entry in the dictionary?
                            //
                            if (!contexts.ContainsKey(interpreter))
                            {
                                //
                                // NOTE: Now, add it to the dictionary of
                                //       test contexts.
                                //
                                contexts.Add(interpreter,
                                    new TestContext(GetCurrentThreadId()));
                            }

                            //
                            // NOTE: Return the test context for the specified
                            //       interpreter.
                            //
                            result = contexts[interpreter];

                            //
                            // NOTE: Save the resulting context for next time.
                            //
                            previousTestContext = result;
                        }
                        else
                        {
                            error = "test contexts not available";
                        }
                    }
                    else
                    {
                        error = "test context slot is invalid";
                    }
                }
                catch (Exception e)
                {
                    error = e;
                }
            }
            else
            {
                error = "invalid interpreter";
            }

            return result;
        }

        ///////////////////////////////////////////////////////////////////////

        public int GetTestContextCount()
        {
            CheckDisposed();

            if (interactiveSlot != null)
            {
                try
                {
                    InterpreterTestContextDictionary contexts =
                        Thread.GetData(testSlot) as
                        InterpreterTestContextDictionary; /* throw */

                    if (contexts != null)
                        return contexts.Count;
                }
                catch
                {
                    // do nothing.
                }
            }

            return 0;
        }

        ///////////////////////////////////////////////////////////////////////

        public bool FreeVariableContext()
        {
            CheckDisposed();

            Result error = null;

            return FreeVariableContext(ref error);
        }

        ///////////////////////////////////////////////////////////////////////

        public bool FreeVariableContext(
            ref Result error
            )
        {
            CheckDisposed();

            if (interpreter != null)
            {
                try
                {
                    if (variableSlot != null)
                    {
                        //
                        // NOTE: Try to obtain the per-interpreter dictionary
                        //       of contexts for this thread.
                        //
                        InterpreterVariableContextDictionary contexts =
                            Thread.GetData(variableSlot) as
                            InterpreterVariableContextDictionary; /* throw */

                        if (contexts != null)
                        {
                            //
                            // NOTE: Does the requested interpreter have an
                            //       entry in the dictionary?
                            //
                            if (contexts.ContainsKey(interpreter))
                                contexts.Remove(interpreter);

                            //
                            // NOTE: If there are no more contexts present,
                            //       free the collection and the thread data
                            //       slot.
                            //
                            if (contexts.Count == 0)
                            {
                                //
                                // NOTE: Release local reference now.
                                //
                                contexts = null;

                                //
                                // NOTE: Clear it in the per-thread data.
                                //
                                Thread.SetData(variableSlot, contexts);
                            }
                        }

                        //
                        // NOTE: Invalidate the cached context for this
                        //       interpreter.
                        //
                        previousVariableContext = null;

                        return true;
                    }
                    else
                    {
                        error = "variable context slot is invalid";
                    }
                }
                catch (Exception e)
                {
                    error = e;
                }
            }
            else
            {
                error = "invalid interpreter";
            }

            return false;
        }

        ///////////////////////////////////////////////////////////////////////

        public IVariableContext GetVariableContext(
            bool create
            )
        {
            CheckDisposed();

            Result error = null;

            return GetVariableContext(create, ref error);
        }

        ///////////////////////////////////////////////////////////////////////

        public IVariableContext GetVariableContext(
            bool create,
            ref Result error
            )
        {
            CheckDisposed();

            IVariableContext result = null;

            if (interpreter != null)
            {
                try
                {
                    result = previousVariableContext;

                    if (CanUseThreadContext(result))
                    {
                        return result;
                    }
                    else if (variableSlot != null)
                    {
                        //
                        // NOTE: Reset the result because it may still contain
                        //       the previous (possibly invalid) context.
                        //
                        result = null;

                        //
                        // NOTE: Try to obtain the per-interpreter dictionary
                        //       of variable contexts for this thread.
                        //
                        InterpreterVariableContextDictionary contexts =
                            Thread.GetData(variableSlot) as
                            InterpreterVariableContextDictionary; /* throw */

                        //
                        // NOTE: Has the dictionary been created yet and/or can
                        //       we create it now?
                        //
                        if (create && (contexts == null))
                        {
                            //
                            // NOTE: Nope, create it now.
                            //
                            contexts = new InterpreterVariableContextDictionary();

                            //
                            // NOTE: Store it in the per-thread data.
                            //
                            Thread.SetData(variableSlot, contexts); /* throw */
                        }

                        //
                        // NOTE: Is the dictionary available now (if not, we
                        //       have been forbidden by the caller from
                        //       automatically creating it).
                        //
                        if (contexts != null)
                        {
                            //
                            // NOTE: Does the requested interpreter have an
                            //       entry in the dictionary?
                            //
                            if (!contexts.ContainsKey(interpreter))
                            {
                                //
                                // NOTE: Create a new call stack for the
                                //       interpreter.
                                //
                                CallStack callStack = new CallStack(
                                    interpreter.RecursionLimit);

                                //
                                // NOTE: If necessary, create a new global call
                                //       frame for the interpreter; all threads
                                //       share the same global call frame for a
                                //       given interpreter.
                                //
                                lock (interpreter.SyncRoot)
                                {
                                    //
                                    // NOTE: Does the interpreter already have
                                    //       a global call frame?  If not,
                                    //       create a new one now.
                                    //
                                    if (interpreter.globalFrame == null)
                                        interpreter.globalFrame =
                                            CallFrameOps.NewGlobal(
                                                interpreter, callStack);

                                    //
                                    // NOTE: Now, add it to the dictionary of
                                    //       variable contexts.
                                    //
                                    contexts.Add(interpreter,
                                        new VariableContext(
                                            GetCurrentThreadId(), callStack,
                                            interpreter.globalFrame, null,
                                            null, null));
                                }

                                //
                                // NOTE: This call frame is never popped.
                                //
                                interpreter.PushGlobalCallFrame(false);
                            }

                            //
                            // NOTE: Return the variable context for the
                            //       specified interpreter.
                            //
                            result = contexts[interpreter];

                            //
                            // NOTE: Save the resulting context for next time.
                            //
                            previousVariableContext = result;
                        }
                        else
                        {
                            error = "variable contexts not available";
                        }
                    }
                    else
                    {
                        error = "variable context slot is invalid";
                    }
                }
                catch (Exception e)
                {
                    error = e;
                }
            }
            else
            {
                error = "invalid interpreter";
            }

            return result;
        }

        ///////////////////////////////////////////////////////////////////////

        public int GetVariableContextCount()
        {
            CheckDisposed();

            try
            {
                if (variableSlot != null)
                {
                    InterpreterVariableContextDictionary contexts =
                        Thread.GetData(variableSlot) as
                        InterpreterVariableContextDictionary; /* throw */

                    if (contexts != null)
                        return contexts.Count;
                }
            }
            catch
            {
                // do nothing.
            }

            return 0;
        }

        ///////////////////////////////////////////////////////////////////////

        public void Free(
            bool global
            )
        {
            IDisposable disposable = GetEngineContext(false) as IDisposable;

            if (disposable != null)
            {
                disposable.Dispose();
                disposable = null;
            }

            FreeEngineContext();

            ///////////////////////////////////////////////////////////////////

            disposable = GetInteractiveContext(false) as IDisposable;

            if (disposable != null)
            {
                disposable.Dispose();
                disposable = null;
            }

            FreeInteractiveContext();

            ///////////////////////////////////////////////////////////////////

            disposable = GetTestContext(false) as IDisposable;

            if (disposable != null)
            {
                disposable.Dispose();
                disposable = null;
            }

            FreeTestContext();

            ///////////////////////////////////////////////////////////////////

            //
            // HACK: *SPECIAL CASE* We cannot dispose the global call frame
            //       unless we are [also] disposing of the interpreter itself;
            //       therefore, use the special Free method here instead of the
            //       Dispose method.  The Free method is guaranteed to do the
            //       right thing with regard to the global call frame (assuming
            //       the "global" paramter is correct).
            //
            IVariableContext variableContext = GetVariableContext(false);

            if (variableContext != null)
            {
                variableContext.Free(global);
                variableContext = null;
            }

            FreeVariableContext();

            ///////////////////////////////////////////////////////////////////

            if (global && (interpreter != null))
                interpreter = null;/* NOT OWNED, DO NOT DISPOSE. */
        }
        #endregion

        ///////////////////////////////////////////////////////////////////////

        #region IDisposable "Pattern" Members
        private bool disposed;
        private void CheckDisposed() /* throw */
        {
#if THROW_ON_DISPOSED
            if (disposed && Engine.IsThrowOnDisposed(interpreter, true))
                throw new ObjectDisposedException(typeof(ContextManager).Name);
#endif
        }

        ///////////////////////////////////////////////////////////////////////

        private /* protected virtual */ void Dispose(
            bool disposing
            )
        {
            if (!disposed)
            {
                if (disposing)
                {
                    ////////////////////////////////////
                    // dispose managed resources here...
                    ////////////////////////////////////

                    Free(true);
                }

                //////////////////////////////////////
                // release unmanaged resources here...
                //////////////////////////////////////

                disposed = true;
            }
        }
        #endregion

        ///////////////////////////////////////////////////////////////////////

        #region IDisposable Members
        public void Dispose()
        {
            Dispose(true);
            GC.SuppressFinalize(this);
        }
        #endregion

        ///////////////////////////////////////////////////////////////////////

        #region Destructor
        ~ContextManager()
        {
            Dispose(false);
        }
        #endregion
    }
}
