/*
 * PlatformOps.cs --
 *
 * Copyright (c) 2007-2012 by Joe Mistachkin.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution of
 * this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 * RCS: @(#) $Id: $
 */

using System;

#if NATIVE && WINDOWS
using System.Diagnostics;
#endif

using System.IO;
using System.Runtime.InteropServices;

#if NATIVE
using System.Security;
#endif

#if !NET_40
using System.Security.Permissions;
#endif

using Eagle._Attributes;
using Eagle._Components.Public;
using Eagle._Constants;
using Eagle._Containers.Public;

namespace Eagle._Components.Private
{
#if NATIVE
#if NET_40
    [SecurityCritical()]
#else
    [SecurityPermission(SecurityAction.LinkDemand, UnmanagedCode = true)]
#endif
#endif
    [ObjectId("751e3aab-6f53-4a0d-bc13-f1ab217ef7dc")]
    internal static class PlatformOps
    {
        private const uint DefaultPageSize = 4096;

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private static readonly StringList Processors = new StringList(new string[] {
            "intel", "mips", "alpha", "ppc", "shx",
            "arm", "ia64", "alpha64", "msil", "amd64",
            "ia32_on_win64"
        });

        private static readonly StringList Platforms = new StringList(new string[] {
            "windows", "windows", "windows",
            "windows", "unix", "windows",
            "unix"
        });

        private static readonly StringList OperatingSystems = new StringList(new string[] {
            "Win32s", "Windows 9x", "Windows NT",
            "Windows CE", "Unix", "Xbox",
            "Darwin"
        });

#if NATIVE && WINDOWS
        private static UnsafeNativeMethods.SYSTEM_INFO systemInfo;
        private static UnsafeNativeMethods.OSVERSIONINFOEX versionInfo;
#endif

        internal static readonly ProcessorArchitecture ProcessorArchitecture =
            ProcessorArchitecture.Unknown;

        internal static readonly OperatingSystemId OperatingSystemId =
            OperatingSystemId.Unknown;

        internal static readonly bool Win32onWin64 = false;

        internal static readonly string Processor = null;
        internal static readonly string Platform = null;
        internal static readonly string OperatingSystem = null;

        internal static readonly uint PageSize = DefaultPageSize;

#if NATIVE
        ///////////////////////////////////////////////////////////////////////////////////////////////
        // Required Native APIs used via P/Invoke
        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region Private Unsafe Native Methods Class
        [SuppressUnmanagedCodeSecurity()]
        [ObjectId("259417ba-e318-4982-b2c0-9f6fd4196b74")]
        private static class UnsafeNativeMethods
        {
#if WINDOWS
            [StructLayout(LayoutKind.Sequential)]
            [ObjectId("a858d038-1313-43d2-a9b0-7a00b2975933")]
            internal struct SYSTEM_INFO
            {
                public ProcessorArchitecture wProcessorArchitecture;
                public ushort wReserved;
                public uint dwPageSize;
                public IntPtr lpMinimumApplicationAddress;
                public IntPtr lpMaximumApplicationAddress;
                public UIntPtr dwActiveProcessorMask;
                public uint dwNumberOfProcessors;
                public uint dwProcessorType;
                public uint dwAllocationGranularity;
                public ushort wProcessorLevel;
                public ushort wProcessorRevision;
            }

            ///////////////////////////////////////////////////////////////////////////////////////////////

            //
            // NOTE: Yes, this has been tested and the size must be exactly 148.
            //
            [StructLayout(LayoutKind.Sequential, CharSet = CharSet.Auto)]
            [ObjectId("f33f3aa7-8ddc-48a0-8ac9-e950dcbaaac1")]
            internal struct OSVERSIONINFOEX
            {
                public uint dwOSVersionInfoSize;
                public uint dwMajorVersion;
                public uint dwMinorVersion;
                public uint dwBuildNumber;
                public OperatingSystemId dwPlatformId;
                [MarshalAs(UnmanagedType.ByValTStr, SizeConst = 128)]
                public string szCSDVersion;
                public ushort wServicePackMajor;
                public ushort wServicePackMinor;
                public short wSuiteMask;
                public byte wProductType;
                public byte wReserved;
            }

            ///////////////////////////////////////////////////////////////////////////////////////////////

            [DllImport(DllName.Kernel32, CallingConvention = CallingConvention.Winapi)]
            internal static extern void GetSystemInfo(ref SYSTEM_INFO systemInfo);

            [DllImport(DllName.Kernel32, CallingConvention = CallingConvention.Winapi, CharSet = CharSet.Auto, BestFitMapping = false, ThrowOnUnmappableChar = true, SetLastError = true)]
            [return: MarshalAs(UnmanagedType.Bool)]
            internal static extern bool GetVersionEx(ref OSVERSIONINFOEX versionInfo);

            [DllImport(DllName.Kernel32, CallingConvention = CallingConvention.Winapi, SetLastError = true)]
            [return: MarshalAs(UnmanagedType.Bool)]
            internal static extern bool IsWow64Process(IntPtr hProcess, [MarshalAs(UnmanagedType.Bool)] ref bool wow64Process);
#endif
        }
        #endregion
#endif

        ///////////////////////////////////////////////////////////////////////////////////////////////

        static PlatformOps()
        {
#if NATIVE && WINDOWS
            if (GetSystemInfo(ref systemInfo))
            {
                //
                // NOTE: What is the processor architecture we are executing on?
                //
                ProcessorArchitecture = systemInfo.wProcessorArchitecture;

                //
                // NOTE: What is the native memory page size?
                //
                PageSize = systemInfo.dwPageSize;
            }
#endif

            //
            // NOTE: What is the name of the processor we are executing on?
            //
            Processor = GetProcessor(ProcessorArchitecture);

#if NATIVE && WINDOWS
            if (GetOsVersionInfo(ref versionInfo))
            {
                //
                // NOTE: What is the platform we are executing on?
                //
                OperatingSystemId = versionInfo.dwPlatformId;
            }
            else
#endif
            {
                OperatingSystemId = (OperatingSystemId)
                    CommonOps.Runtime.GetOperatingSystemPlatform();
            }

            //
            // NOTE: Check if this process is running as Win32-on-Win64
            //       (WoW64).
            //
#if NATIVE && WINDOWS
            Win32onWin64 = IsWin32onWin64();
#else
            Win32onWin64 = false;
#endif

            //
            // NOTE: What is the name of the platform we are executing on?
            //
            if (OperatingSystemId != OperatingSystemId.Unknown)
                Platform = GetPlatform(OperatingSystemId);
            else
                Platform = GetPlatform((OperatingSystemId)
                    CommonOps.Runtime.GetOperatingSystemPlatform());

            //
            // NOTE: What is the name of the operating system we are executing on?
            //
            OperatingSystem = GetOperatingSystem(OperatingSystemId);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

#if NATIVE && WINDOWS
        private static bool GetSystemInfo(
            ref UnsafeNativeMethods.SYSTEM_INFO systemInfo
            )
        {
            try
            {
                /* CANNOT FAIL? */
                UnsafeNativeMethods.GetSystemInfo(ref systemInfo);

                return true;
            }
            catch
            {
                // do nothing.
            }

            return false;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private static bool GetOsVersionInfo(
            ref UnsafeNativeMethods.OSVERSIONINFOEX versionInfo
            )
        {
            try
            {
                versionInfo.dwOSVersionInfoSize =
                    (uint)Marshal.SizeOf(versionInfo);

                return UnsafeNativeMethods.GetVersionEx(ref versionInfo);
            }
            catch
            {
                // do nothing.
            }

            return false;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private static bool IsWin32onWin64()
        {
            try
            {
                Process process = Process.GetCurrentProcess();

                if (process != null)
                {
                    bool wow64Process = false;

                    if (UnsafeNativeMethods.IsWow64Process(
                            process.Handle, ref wow64Process))
                    {
                        return wow64Process;
                    }
                }
            }
            catch
            {
                // do nothing.
            }

            return false;
        }
#endif

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public static bool IsUnixOperatingSystem()
        {
            return ((OperatingSystemId == OperatingSystemId.Unix) ||
                (OperatingSystemId == OperatingSystemId.Darwin) ||
                (OperatingSystemId == OperatingSystemId.Mono_on_Unix));
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public static bool IsWindowsOperatingSystem()
        {
            return ((OperatingSystemId == OperatingSystemId.Windows9x) ||
                (OperatingSystemId == OperatingSystemId.WindowsNT));
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public static bool Is64BitOperatingSystem()
        {
            return (IntPtr.Size == sizeof(ulong));
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public static bool Is32BitOperatingSystem()
        {
            return (IntPtr.Size == sizeof(uint));
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public static int GetOperatingSystemBits()
        {
            return (IntPtr.Size * ConversionOps.ByteBits); // (e.g. 32, 64, etc)
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private static string GetProcessor(
            ProcessorArchitecture processorArchitecture
            )
        {
            if ((processorArchitecture >= 0) &&
                (processorArchitecture < (ProcessorArchitecture)Processors.Count))
            {
                return Processors[(int)processorArchitecture];
            }

            return "unknown";
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private static string GetPlatform(
            OperatingSystemId platformId
            )
        {
            if ((platformId >= 0) &&
                (platformId < (OperatingSystemId)Platforms.Count))
            {
                return Platforms[(int)platformId];
            } else if (platformId == OperatingSystemId.Mono_on_Unix)
            {
                return OperatingSystemId.Unix.ToString();
            }

            return "unknown";
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private static string GetOperatingSystem(
            OperatingSystemId platformId
            )
        {
            if ((platformId >= 0) &&
                (platformId < (OperatingSystemId)OperatingSystems.Count))
            {
                return OperatingSystems[(int)platformId];
            }
            else if (platformId == OperatingSystemId.Mono_on_Unix)
            {
                return platformId.ToString().Replace(Characters.Underscore, Characters.Space);
            }

            return "unknown";
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public static string GetServicePackString()
        {
#if NATIVE && WINDOWS
            return FormatOps.MajorMinor(new Version(
                versionInfo.wServicePackMajor, versionInfo.wServicePackMinor));
#else
            return Environment.OSVersion.ServicePack;
#endif
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public static string GetUserName()
        {
            return Environment.UserDomainName + Path.DirectorySeparatorChar + Environment.UserName;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

#if NATIVE && WINDOWS
        public static bool CheckVersion(
            PlatformID platformId,
            int major,
            int minor,
            short servicePackMajor,
            short servicePackMinor
            )
        {
            if (CommonOps.Runtime.GetOperatingSystemPlatform() == platformId)
            {
                Version osVersion = CommonOps.Runtime.GetOperatingSystemVersion();

                if (osVersion.Major > major)
                {
                    return true;
                }
                else if ((osVersion.Major == major) && (osVersion.Minor > minor))
                {
                    return true;
                }
                else if ((osVersion.Major == major) && (osVersion.Minor == minor))
                {
                    ushort osServicePackMajor = versionInfo.wServicePackMajor;
                    ushort osServicePackMinor = versionInfo.wServicePackMinor;

                    if (osServicePackMajor > servicePackMajor)
                    {
                        return true;
                    }
                    else if ((osServicePackMajor == servicePackMajor) &&
                        (osServicePackMinor >= servicePackMinor))
                    {
                        return true;
                    }
                }
            }

            return false;
        }
#endif
    }
}
